import inspect
from collections.abc import Callable, Iterable
from functools import wraps
from typing import Any, TypeVar, cast

R = TypeVar("R")


class AdaptError(TypeError): ...


def adapt_with_mapper[R](
    func: Callable[..., R], mapper: Callable[..., Any]
) -> Callable[..., R]:
    """
    Transform a function to accept a different input type by using a mapper function.

    Instead of calling `greet(user_id)`, you can call `greet(context)` where the
    `user_id` gets automatically extracted from the `context`.
    """

    func_sig = inspect.signature(func)
    mapper_sig = inspect.signature(mapper)

    mapper_return_type = mapper_sig.return_annotation

    if mapper_return_type is inspect._empty:
        raise AdaptError("mapper must have a return type annotation")

    any_replacements = False
    mapper_argument_types = set(
        param.annotation for param in mapper_sig.parameters.values()
    )
    if inspect._empty in mapper_argument_types:
        raise AdaptError("mapper must have type-annotated parameters")

    new_params = list(mapper_sig.parameters.values())
    for func_sig_param in func_sig.parameters.values():
        if func_sig_param.annotation == mapper_return_type:
            any_replacements = True
        elif func_sig_param.annotation not in mapper_argument_types:
            new_params.append(func_sig_param)

    if not any_replacements:
        return func

    new_sig = func_sig.replace(parameters=new_params)

    def get_annotations(sig: inspect.Signature) -> dict[str, Any]:
        annotations = {}
        annotations["return"] = sig.return_annotation
        for param in sig.parameters.values():
            if param.annotation is not inspect._empty:
                annotations[param.name] = param.annotation
        return annotations

    def bind_args(*args, **kwargs) -> inspect.BoundArguments:
        bound_args = new_sig.bind(*args, **kwargs)
        bound_args.apply_defaults()
        return bound_args

    def invoke_mapper(bound_args: inspect.BoundArguments) -> Any:
        mapper_args = {}
        for mapper_param in mapper_sig.parameters.values():
            mapper_args[mapper_param.name] = bound_args.arguments[mapper_param.name]
        return mapper(**mapper_args)

    def invoke_func(bound_args: inspect.BoundArguments, mapped_value: Any) -> Any:
        func_args = {}
        for func_param in func_sig.parameters.values():
            if func_param.annotation == mapper_return_type:
                func_args[func_param.name] = mapped_value
            else:
                func_args[func_param.name] = bound_args.arguments[func_param.name]
        return func(**func_args)

    if inspect.iscoroutinefunction(func):

        @wraps(func)
        async def awrapper(*args: Any, **kwargs: Any) -> Any:
            bound_args = bind_args(*args, **kwargs)
            if inspect.iscoroutinefunction(mapper):
                mapped_value = await invoke_mapper(bound_args)
            else:
                mapped_value = invoke_mapper(bound_args)
            return await invoke_func(bound_args, mapped_value)

        awrapper.__signature__ = new_sig  # type: ignore[attr-defined]
        awrapper.__annotations__ = get_annotations(new_sig)
        return cast(Callable[..., R], awrapper)

    else:
        if inspect.iscoroutinefunction(mapper):
            raise AdaptError("Async mapper used with sync function")

        @wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> R:
            bound_args = bind_args(*args, **kwargs)
            mapped_value = invoke_mapper(bound_args)
            return invoke_func(bound_args, mapped_value)

        wrapper.__signature__ = new_sig  # type: ignore[attr-defined]
        wrapper.__annotations__ = get_annotations(new_sig)

        return wrapper


def adapt_with_mappers[R](
    func: Callable[..., R],
    mappers: Iterable[Callable[..., Any]],
) -> Callable[..., R]:
    for mapper in mappers:
        func = adapt_with_mapper(func, mapper)
    return func



import asyncio
import textwrap
from enum import StrEnum
from typing import Any, ClassVar, Literal, TypedDict

import httpx
from pydantic import BaseModel, ConfigDict, Field

from dbt_mcp.config.config_providers import ConfigProvider, DiscoveryConfig
from dbt_mcp.discovery.graphql import load_query
from dbt_mcp.errors import InvalidParameterError, ToolCallError
from dbt_mcp.gql.errors import raise_gql_error

DEFAULT_PAGE_SIZE = 100
DEFAULT_MAX_NODE_QUERY_LIMIT = 10000


class GraphQLQueries:
    GET_MODELS = textwrap.dedent("""
        query GetModels(
            $environmentId: BigInt!,
            $modelsFilter: ModelAppliedFilter,
            $after: String,
            $first: Int,
            $sort: AppliedModelSort
        ) {
            environment(id: $environmentId) {
                applied {
                    models(filter: $modelsFilter, after: $after, first: $first, sort: $sort) {
                        pageInfo {
                            endCursor
                        }
                        edges {
                            node {
                                name
                                uniqueId
                                description
                            }
                        }
                    }
                }
            }
        }
    """)

    GET_MODEL_HEALTH = textwrap.dedent("""
        query GetModelDetails(
            $environmentId: BigInt!,
            $modelsFilter: ModelAppliedFilter
            $first: Int,
        ) {
            environment(id: $environmentId) {
                applied {
                    models(filter: $modelsFilter, first: $first) {
                        edges {
                            node {
                                name
                                uniqueId
                                executionInfo {
                                    lastRunGeneratedAt
                                    lastRunStatus
                                    executeCompletedAt
                                    executeStartedAt
                                }
                                tests {
                                    name
                                    description
                                    columnName
                                    testType
                                    executionInfo {
                                        lastRunGeneratedAt
                                        lastRunStatus
                                        executeCompletedAt
                                        executeStartedAt
                                    }
                                }
                                ancestors(types: [Model, Source, Seed, Snapshot]) {
                                  ... on ModelAppliedStateNestedNode {
                                    name
                                    uniqueId
                                    resourceType
                                    materializedType
                                    modelexecutionInfo: executionInfo {
                                      lastRunStatus
                                      executeCompletedAt
                                      }
                                  }
                                  ... on SnapshotAppliedStateNestedNode {
                                    name
                                    uniqueId
                                    resourceType
                                    snapshotExecutionInfo: executionInfo {
                                      lastRunStatus
                                      executeCompletedAt
                                    }
                                  }
                                  ... on SeedAppliedStateNestedNode {
                                    name
                                    uniqueId
                                    resourceType
                                    seedExecutionInfo: executionInfo {
                                      lastRunStatus
                                      executeCompletedAt
                                    }
                                  }
                                  ... on SourceAppliedStateNestedNode {
                                    sourceName
                                    name
                                    resourceType
                                    freshness {
                                      maxLoadedAt
                                      maxLoadedAtTimeAgoInS
                                      freshnessStatus
                                    }
                                  }
                              }
                            }
                        }
                    }
                }
            }
        }
    """)

    COMMON_FIELDS_PARENTS_CHILDREN = textwrap.dedent("""
        {
        ... on ExposureAppliedStateNestedNode {
            resourceType
            name
            description
        }
        ... on ExternalModelNode {
            resourceType
            description
            name
        }
        ... on MacroDefinitionNestedNode {
            resourceType
            name
            description
        }
        ... on MetricDefinitionNestedNode {
            resourceType
            name
            description
        }
        ... on ModelAppliedStateNestedNode {
            resourceType
            name
            description
        }
        ... on SavedQueryDefinitionNestedNode {
            resourceType
            name
            description
        }
        ... on SeedAppliedStateNestedNode {
            resourceType
            name
            description
        }
        ... on SemanticModelDefinitionNestedNode {
            resourceType
            name
            description
        }
        ... on SnapshotAppliedStateNestedNode {
            resourceType
            name
            description
        }
        ... on SourceAppliedStateNestedNode {
            resourceType
            sourceName
            uniqueId
            name
            description
        }
        ... on TestAppliedStateNestedNode {
            resourceType
            name
            description
        }
    """)

    GET_MODEL_PARENTS = (
        textwrap.dedent("""
        query GetModelParents(
            $environmentId: BigInt!,
            $modelsFilter: ModelAppliedFilter
            $first: Int,
        ) {
            environment(id: $environmentId) {
                applied {
                    models(filter: $modelsFilter, first: $first) {
                        pageInfo {
                            endCursor
                        }
                        edges {
                            node {
                                parents
    """)
        + COMMON_FIELDS_PARENTS_CHILDREN
        + textwrap.dedent("""
                                }
                            }
                        }
                    }
                }
            }
        }
    """)
    )

    GET_MODEL_CHILDREN = (
        textwrap.dedent("""
        query GetModelChildren(
            $environmentId: BigInt!,
            $modelsFilter: ModelAppliedFilter
            $first: Int,
        ) {
            environment(id: $environmentId) {
                applied {
                    models(filter: $modelsFilter, first: $first) {
                        pageInfo {
                            endCursor
                        }
                        edges {
                            node {
                                children
    """)
        + COMMON_FIELDS_PARENTS_CHILDREN
        + textwrap.dedent("""
                                }
                            }
                        }
                    }
                }
            }
        }
    """)
    )

    GET_SOURCES = textwrap.dedent("""
        query GetSources(
            $environmentId: BigInt!,
            $sourcesFilter: SourceAppliedFilter,
            $after: String,
            $first: Int
        ) {
            environment(id: $environmentId) {
                applied {
                    sources(filter: $sourcesFilter, after: $after, first: $first) {
                        pageInfo {
                            hasNextPage
                            endCursor
                        }
                        edges {
                            node {
                                name
                                uniqueId
                                identifier
                                description
                                sourceName
                                resourceType
                                database
                                schema
                                freshness {
                                    maxLoadedAt
                                    maxLoadedAtTimeAgoInS
                                    freshnessStatus
                                }
                            }
                        }
                    }
                }
            }
        }
    """)

    GET_EXPOSURES = textwrap.dedent("""
        query Exposures($environmentId: BigInt!, $first: Int, $after: String) {
            environment(id: $environmentId) {
                definition {
                    exposures(first: $first, after: $after) {
                        totalCount
                        pageInfo {
                            hasNextPage
                            endCursor
                        }
                        edges {
                            node {
                                name
                                uniqueId
                                url
                                description
                            }
                        }
                    }
                }
            }
        }
    """)

    # Lineage query
    GET_FULL_LINEAGE = load_query("get_full_lineage.gql")


class MetadataAPIClient:
    def __init__(self, config_provider: ConfigProvider[DiscoveryConfig]):
        self.config_provider = config_provider

    async def execute_query(self, query: str, variables: dict) -> dict:
        config = await self.config_provider.get_config()
        url = config.url
        headers = config.headers_provider.get_headers()

        async with httpx.AsyncClient() as client:
            response = await client.post(
                url=url,
                json={"query": query, "variables": variables},
                headers=headers,
            )
            response.raise_for_status()
            return response.json()


class PageInfo(BaseModel):
    model_config = ConfigDict(extra="ignore", populate_by_name=True)

    end_cursor: str | None = Field(default=None, alias="endCursor")
    has_next_page: bool | None = Field(default=None, alias="hasNextPage")


class PaginatedResourceFetcher:
    def __init__(
        self,
        api_client: MetadataAPIClient,
        *,
        edges_path: tuple[str, ...],
        page_info_path: tuple[str, ...],
        page_size: int = DEFAULT_PAGE_SIZE,
        max_node_query_limit: int = DEFAULT_MAX_NODE_QUERY_LIMIT,
    ):
        self.api_client = api_client
        self._edges_path = edges_path
        self._page_info_path = page_info_path
        self._page_size = page_size
        self._max_node_query_limit = max_node_query_limit

    async def get_environment_id(self) -> int:
        config = await self.api_client.config_provider.get_config()
        return config.environment_id

    def _extract_path(self, payload: dict, path: tuple[str, ...]) -> Any:
        current = payload
        for key in path:
            current = current[key]
        return current

    def _parse_edges(self, result: dict) -> list[dict]:
        raise_gql_error(result)
        edges = self._extract_path(result, self._edges_path)
        parsed_edges: list[dict] = []
        if not edges:
            return parsed_edges
        for edge in edges:
            if not isinstance(edge, dict) or "node" not in edge:
                continue
            node = edge["node"]
            if not isinstance(node, dict):
                continue
            parsed_edges.append(node)
        return parsed_edges

    def _should_continue(
        self,
        page_info: PageInfo,
        previous_cursor: str | None,
    ) -> bool:
        next_cursor = page_info.end_cursor
        has_next = page_info.has_next_page
        next_cursor_valid = bool(next_cursor) and next_cursor != previous_cursor
        if isinstance(has_next, bool):
            return has_next and next_cursor_valid
        return next_cursor_valid

    async def fetch_paginated(
        self,
        query: str,
        variables: dict[str, Any],
    ) -> list[dict]:
        environment_id = await self.get_environment_id()
        collected: list[dict] = []
        current_cursor: str | None = None
        while True:
            if len(collected) >= self._max_node_query_limit:
                break
            remaining_capacity = self._max_node_query_limit - len(collected)
            request_variables = variables.copy()
            request_variables["environmentId"] = environment_id
            request_variables["first"] = min(self._page_size, remaining_capacity)
            if current_cursor is not None:
                request_variables["after"] = current_cursor
            result = await self.api_client.execute_query(query, request_variables)
            page_edges = self._parse_edges(result)
            collected.extend(page_edges)
            page_info_data = self._extract_path(result, self._page_info_path)
            page_info = PageInfo(**page_info_data)
            previous_cursor = current_cursor
            current_cursor = page_info.end_cursor
            if not self._should_continue(page_info, previous_cursor):
                break
        return collected


class ModelFilter(TypedDict, total=False):
    modelingLayer: Literal["marts"] | None


class SourceFilter(TypedDict, total=False):
    sourceNames: list[str]
    uniqueIds: list[str] | None
    identifier: str


class ModelsFetcher:
    def __init__(
        self,
        api_client: MetadataAPIClient,
        paginator: PaginatedResourceFetcher,
    ):
        self.api_client = api_client
        self._paginator = paginator

    def _get_model_filters(
        self, model_name: str | None = None, unique_id: str | None = None
    ) -> dict[str, list[str] | str]:
        if unique_id:
            return {"uniqueIds": [unique_id]}
        elif model_name:
            return {"identifier": model_name}
        else:
            raise InvalidParameterError(
                "Either model_name or unique_id must be provided"
            )

    async def fetch_models(self, model_filter: ModelFilter | None = None) -> list[dict]:
        return await self._paginator.fetch_paginated(
            GraphQLQueries.GET_MODELS,
            variables={
                "modelsFilter": model_filter or {},
                "sort": {"field": "queryUsageCount", "direction": "desc"},
            },
        )

    async def fetch_model_parents(
        self, model_name: str | None = None, unique_id: str | None = None
    ) -> list[dict]:
        model_filters = self._get_model_filters(model_name, unique_id)
        variables = {
            "environmentId": await self._paginator.get_environment_id(),
            "modelsFilter": model_filters,
            "first": 1,
        }
        result = await self.api_client.execute_query(
            GraphQLQueries.GET_MODEL_PARENTS, variables
        )
        raise_gql_error(result)
        edges = result["data"]["environment"]["applied"]["models"]["edges"]
        if not edges:
            return []
        return edges[0]["node"]["parents"]

    async def fetch_model_children(
        self, model_name: str | None = None, unique_id: str | None = None
    ) -> list[dict]:
        model_filters = self._get_model_filters(model_name, unique_id)
        variables = {
            "environmentId": await self._paginator.get_environment_id(),
            "modelsFilter": model_filters,
            "first": 1,
        }
        result = await self.api_client.execute_query(
            GraphQLQueries.GET_MODEL_CHILDREN, variables
        )
        raise_gql_error(result)
        edges = result["data"]["environment"]["applied"]["models"]["edges"]
        if not edges:
            return []
        return edges[0]["node"]["children"]

    async def fetch_model_health(
        self, model_name: str | None = None, unique_id: str | None = None
    ) -> list[dict]:
        model_filters = self._get_model_filters(model_name, unique_id)
        variables = {
            "environmentId": await self._paginator.get_environment_id(),
            "modelsFilter": model_filters,
            "first": 1,
        }
        result = await self.api_client.execute_query(
            GraphQLQueries.GET_MODEL_HEALTH, variables
        )
        raise_gql_error(result)
        edges = result["data"]["environment"]["applied"]["models"]["edges"]
        if not edges:
            return []
        return edges[0]["node"]


class ExposuresFetcher:
    def __init__(
        self,
        api_client: MetadataAPIClient,
        paginator: PaginatedResourceFetcher,
    ):
        self.api_client = api_client
        self._paginator = paginator

    async def fetch_exposures(self) -> list[dict]:
        return await self._paginator.fetch_paginated(
            GraphQLQueries.GET_EXPOSURES,
            variables={},
        )


class SourcesFetcher:
    def __init__(
        self,
        api_client: MetadataAPIClient,
        paginator: PaginatedResourceFetcher,
    ):
        self.api_client = api_client
        self._paginator = paginator

    async def get_environment_id(self) -> int:
        return await self._paginator.get_environment_id()

    async def fetch_sources(
        self,
        source_names: list[str] | None = None,
        unique_ids: list[str] | None = None,
    ) -> list[dict]:
        source_filter: SourceFilter = {}
        if source_names is not None:
            source_filter["sourceNames"] = source_names
        if unique_ids is not None:
            source_filter["uniqueIds"] = unique_ids

        return await self._paginator.fetch_paginated(
            GraphQLQueries.GET_SOURCES,
            variables={"sourcesFilter": source_filter},
        )


class AppliedResourceType(StrEnum):
    MODEL = "model"
    SOURCE = "source"
    EXPOSURE = "exposure"
    TEST = "test"
    SEED = "seed"
    SNAPSHOT = "snapshot"
    MACRO = "macro"
    SEMANTIC_MODEL = "semantic_model"


class ResourceDetailsFetcher:
    GET_PACKAGES_QUERY = load_query("get_packages.gql")
    GET_MODELS_DETAILS_QUERY = load_query("get_model_details.gql")
    GET_SOURCES_QUERY = load_query("get_source_details.gql")
    GET_EXPOSURES_QUERY = load_query("get_exposure_details.gql")
    GET_TESTS_QUERY = load_query("get_test_details.gql")
    GET_SEEDS_QUERY = load_query("get_seed_details.gql")
    GET_SNAPSHOTS_QUERY = load_query("get_snapshot_details.gql")
    GET_MACROS_QUERY = load_query("get_macro_details.gql")
    GET_SEMANTIC_MODELS_QUERY = load_query("get_semantic_model_details.gql")

    GQL_QUERIES: ClassVar[dict[AppliedResourceType, str]] = {
        AppliedResourceType.MODEL: GET_MODELS_DETAILS_QUERY,
        AppliedResourceType.SOURCE: GET_SOURCES_QUERY,
        AppliedResourceType.EXPOSURE: GET_EXPOSURES_QUERY,
        AppliedResourceType.TEST: GET_TESTS_QUERY,
        AppliedResourceType.SEED: GET_SEEDS_QUERY,
        AppliedResourceType.SNAPSHOT: GET_SNAPSHOTS_QUERY,
        AppliedResourceType.MACRO: GET_MACROS_QUERY,
        AppliedResourceType.SEMANTIC_MODEL: GET_SEMANTIC_MODELS_QUERY,
    }

    RESOURCE_TYPE_TO_GQL_TYPE: ClassVar[dict[AppliedResourceType, str]] = {
        AppliedResourceType.MODEL: "Model",
        AppliedResourceType.SOURCE: "Source",
        AppliedResourceType.EXPOSURE: "Exposure",
        AppliedResourceType.TEST: "Test",
        AppliedResourceType.SEED: "Seed",
        AppliedResourceType.SNAPSHOT: "Snapshot",
        AppliedResourceType.MACRO: "Macro",
        AppliedResourceType.SEMANTIC_MODEL: "SemanticModel",
    }

    def __init__(
        self,
        api_client: MetadataAPIClient,
    ):
        self.api_client = api_client

    async def fetch_details(
        self,
        resource_type: AppliedResourceType,
        name: str | None = None,
        unique_id: str | None = None,
    ) -> list[dict]:
        normalized_name = name.strip().lower() if name else None
        normalized_unique_id = unique_id.strip().lower() if unique_id else None
        environment_id = (
            await self.api_client.config_provider.get_config()
        ).environment_id
        if not normalized_name and not normalized_unique_id:
            raise InvalidParameterError("Either name or unique_id must be provided")
        if (
            normalized_name
            and normalized_unique_id
            and normalized_name != normalized_unique_id.split(".")[-1]
        ):
            raise InvalidParameterError(
                f"Name and unique_id do not match. The unique_id does not end with {normalized_name}."
            )
        if not normalized_unique_id:
            assert normalized_name is not None, "Name must be provided"
            packages_result = await asyncio.gather(
                self.api_client.execute_query(
                    self.GET_PACKAGES_QUERY,
                    variables={"resource": "macro", "environmentId": environment_id},
                ),
                self.api_client.execute_query(
                    self.GET_PACKAGES_QUERY,
                    variables={"resource": "model", "environmentId": environment_id},
                ),
            )
            raise_gql_error(packages_result[0])
            raise_gql_error(packages_result[1])
            macro_packages = packages_result[0]["data"]["environment"]["applied"][
                "packages"
            ]
            model_packages = packages_result[1]["data"]["environment"]["applied"][
                "packages"
            ]
            if not macro_packages and not model_packages:
                raise InvalidParameterError("No packages found for project")
            unique_ids = [
                f"{resource_type.value.lower()}.{package_name}.{normalized_name}"
                for package_name in macro_packages + model_packages
            ]
        else:
            unique_ids = [normalized_unique_id]
        query = self.GQL_QUERIES[resource_type]
        variables = {
            "environmentId": environment_id,
            "filter": {
                "uniqueIds": unique_ids,
                "types": [self.RESOURCE_TYPE_TO_GQL_TYPE[resource_type]],
            },
            "first": len(unique_ids),
        }
        get_details_result = await self.api_client.execute_query(query, variables)
        raise_gql_error(get_details_result)
        edges = get_details_result["data"]["environment"]["applied"]["resources"][
            "edges"
        ]
        if not edges:
            return []
        return [e["node"] for e in edges]


class LineageResourceType(StrEnum):
    """Resource types supported by the lineage API."""

    MODEL = "Model"
    SOURCE = "Source"
    SEED = "Seed"
    SNAPSHOT = "Snapshot"
    EXPOSURE = "Exposure"
    METRIC = "Metric"
    SEMANTIC_MODEL = "SemanticModel"
    SAVED_QUERY = "SavedQuery"
    TEST = "Test"


class LineageFetcher:
    """Fetcher for lineage data. Returns nodes connected to the target."""

    def __init__(self, api_client: MetadataAPIClient):
        self.api_client = api_client

    async def fetch_lineage(
        self,
        unique_id: str,
        depth: int,
        types: list[LineageResourceType] | None = None,
    ) -> list[dict]:
        """Fetch lineage graph filtered to nodes connected to unique_id.

        Args:
            unique_id: The dbt unique ID of the resource to get lineage for.
            types: List of resource types to include. If None, includes all types.

        Returns:
            List of nodes connected to unique_id (upstream + downstream).
        """
        if depth <= 0:
            raise ToolCallError("Depth must be greater than 0")
        config = await self.api_client.config_provider.get_config()
        type_filter = [
            t.value for t in (types if types is not None else LineageResourceType)
        ]
        variables = {
            "environmentId": config.environment_id,
            "types": type_filter,
            # uniqueId removed - not used by GraphQL
        }

        result = await self.api_client.execute_query(
            GraphQLQueries.GET_FULL_LINEAGE, variables
        )
        raise_gql_error(result)

        all_nodes = (
            result.get("data", {})
            .get("environment", {})
            .get("applied", {})
            .get("lineage", [])
        )

        # Filter to connected nodes only
        return self._filter_connected_nodes(all_nodes, unique_id, depth)

    def _filter_connected_nodes(
        self, nodes: list[dict], target_id: str, depth: int
    ) -> list[dict]:
        """Return only nodes connected to target_id (upstream and downstream).

        Uses BFS to find all nodes reachable from target in both directions.
        """
        node_map = {
            n["uniqueId"]: n
            for n in nodes
            if (resource_type := n.get("resourceType"))
            and isinstance(resource_type, str)
            # Filtering out macros because they have large
            # dependency graphs that aren't always useful.
            and resource_type.strip().lower() != "macro"
        }

        if target_id not in node_map:
            return []

        # BFS to find all connected nodes
        connected = {target_id}
        queue = [(target_id, 0)]

        while queue:
            current_id, current_depth = queue.pop(0)
            node = node_map.get(current_id)
            if not node:
                continue

            # Stop traversing beyond the depth limit
            if current_depth >= depth:
                continue

            # Traverse upstream (parents)
            for parent_id in node.get("parentIds", []):
                if parent_id not in connected and parent_id in node_map:
                    connected.add(parent_id)
                    queue.append((parent_id, current_depth + 1))

            # Traverse downstream (children)
            for candidate in nodes:
                candidate_id = candidate.get("uniqueId")
                if not candidate_id or candidate_id not in node_map:
                    continue
                if (
                    current_id in candidate.get("parentIds", [])
                    and candidate_id not in connected
                ):
                    connected.add(candidate_id)
                    queue.append((candidate_id, current_depth + 1))

        # Return in original order
        return [node_map[uid] for uid in connected]


class ModelPerformanceFetcher:
    """Fetches model execution performance data from Discovery API."""

    GET_MODEL_PERFORMANCE_QUERY = load_query("get_model_performance.gql")

    def __init__(
        self,
        api_client: MetadataAPIClient,
        resource_details_fetcher: ResourceDetailsFetcher,
    ):
        self.api_client = api_client
        self._resource_details_fetcher = resource_details_fetcher

    async def fetch_performance(
        self,
        name: str | None = None,
        unique_id: str | None = None,
        num_runs: int = 1,
        include_tests: bool = False,
    ) -> list[dict]:
        """Fetch model performance data.

        Args:
            name: Model name (resolved to unique_id if provided)
            unique_id: Fully-qualified unique ID (preferred)
            num_runs: Number of historical runs to retrieve
            include_tests: If True, include test execution history for each run

        Returns:
            List of performance records sorted by executeStartedAt descending

        Raises:
            InvalidParameterError: If neither name nor unique_id provided
            ToolCallError: If model not found or API error
        """
        if not name and not unique_id:
            raise InvalidParameterError("Either 'name' or 'unique_id' must be provided")

        # Resolve name to unique_id if needed
        resolved_unique_id = unique_id
        if not resolved_unique_id:
            # Re-use resource_details_fetcher from ResourceDetailsFetcher to resolve name
            details = await self._resource_details_fetcher.fetch_details(
                resource_type=AppliedResourceType.MODEL,
                name=name,
            )
            if not details:
                raise ToolCallError(f"Model not found: {name}")
            # Model name can map to multiple unique_ids - require disambiguation
            # For example, if multiple dbt packages define a model with the same name
            if len(details) > 1:
                matches = ", ".join(
                    sorted(d.get("uniqueId", "") for d in details if d.get("uniqueId"))
                )
                raise ToolCallError(
                    f"Multiple models found for name '{name}'. "
                    "Please provide the unique_id instead. "
                    f"Matches: {matches}"
                )
            resolved_unique_id = details[0]["uniqueId"]

        config = await self.api_client.config_provider.get_config()
        variables = {
            "environmentId": config.environment_id,
            "uniqueId": resolved_unique_id,
            "lastRunCount": num_runs,
        }

        result = await self.api_client.execute_query(
            self.GET_MODEL_PERFORMANCE_QUERY,
            variables,
        )
        raise_gql_error(result)

        runs = (
            result.get("data", {})
            .get("environment", {})
            .get("applied", {})
            .get("modelHistoricalRuns", [])
        )

        if not include_tests and runs:
            for run in runs:
                run.pop("tests", None)

        return runs or []



import logging
from dataclasses import dataclass

from dbtsl.api.shared.query_params import GroupByParam
from mcp.server.fastmcp import FastMCP

from dbt_mcp.config.config_providers import ConfigProvider, SemanticLayerConfig
from dbt_mcp.prompts.prompts import get_prompt
from dbt_mcp.semantic_layer.client import (
    SemanticLayerClientProvider,
    SemanticLayerFetcher,
)
from dbt_mcp.semantic_layer.types import (
    DimensionToolResponse,
    EntityToolResponse,
    GetMetricsCompiledSqlSuccess,
    MetricToolResponse,
    OrderByParam,
    QueryMetricsSuccess,
    SavedQueryToolResponse,
)
from dbt_mcp.tools.definitions import dbt_mcp_tool
from dbt_mcp.tools.register import register_tools
from dbt_mcp.tools.tool_names import ToolName
from dbt_mcp.tools.toolsets import Toolset

logger = logging.getLogger(__name__)


@dataclass
class SemanticLayerToolContext:
    semantic_layer_fetcher: SemanticLayerFetcher

    def __init__(
        self,
        config_provider: ConfigProvider[SemanticLayerConfig],
        client_provider: SemanticLayerClientProvider,
    ):
        self.semantic_layer_fetcher = SemanticLayerFetcher(
            config_provider=config_provider, client_provider=client_provider
        )


@dbt_mcp_tool(
    description=get_prompt("semantic_layer/list_metrics"),
    title="List Metrics",
    read_only_hint=True,
    destructive_hint=False,
    idempotent_hint=True,
)
async def list_metrics(
    context: SemanticLayerToolContext, search: str | None = None
) -> list[MetricToolResponse]:
    return await context.semantic_layer_fetcher.list_metrics(search=search)


@dbt_mcp_tool(
    description=get_prompt("semantic_layer/list_saved_queries"),
    title="List Saved Queries",
    read_only_hint=True,
    destructive_hint=False,
    idempotent_hint=True,
)
async def list_saved_queries(
    context: SemanticLayerToolContext,
    search: str | None = None,
) -> list[SavedQueryToolResponse]:
    return await context.semantic_layer_fetcher.list_saved_queries(search=search)


@dbt_mcp_tool(
    description=get_prompt("semantic_layer/get_dimensions"),
    title="Get Dimensions",
    read_only_hint=True,
    destructive_hint=False,
    idempotent_hint=True,
)
async def get_dimensions(
    context: SemanticLayerToolContext, metrics: list[str], search: str | None = None
) -> list[DimensionToolResponse]:
    return await context.semantic_layer_fetcher.get_dimensions(
        metrics=metrics, search=search
    )


@dbt_mcp_tool(
    description=get_prompt("semantic_layer/get_entities"),
    title="Get Entities",
    read_only_hint=True,
    destructive_hint=False,
    idempotent_hint=True,
)
async def get_entities(
    context: SemanticLayerToolContext, metrics: list[str], search: str | None = None
) -> list[EntityToolResponse]:
    return await context.semantic_layer_fetcher.get_entities(
        metrics=metrics, search=search
    )


@dbt_mcp_tool(
    description=get_prompt("semantic_layer/query_metrics"),
    title="Query Metrics",
    read_only_hint=True,
    destructive_hint=False,
    idempotent_hint=True,
)
async def query_metrics(
    context: SemanticLayerToolContext,
    metrics: list[str],
    group_by: list[GroupByParam] | None = None,
    order_by: list[OrderByParam] | None = None,
    where: str | None = None,
    limit: int | None = None,
) -> str:
    result = await context.semantic_layer_fetcher.query_metrics(
        metrics=metrics,
        group_by=group_by,
        order_by=order_by,
        where=where,
        limit=limit,
    )
    if isinstance(result, QueryMetricsSuccess):
        return result.result
    else:
        return result.error


@dbt_mcp_tool(
    description=get_prompt("semantic_layer/get_metrics_compiled_sql"),
    title="Compile SQL",
    read_only_hint=True,
    destructive_hint=False,
    idempotent_hint=True,
)
async def get_metrics_compiled_sql(
    context: SemanticLayerToolContext,
    metrics: list[str],
    group_by: list[GroupByParam] | None = None,
    order_by: list[OrderByParam] | None = None,
    where: str | None = None,
    limit: int | None = None,
) -> str:
    result = await context.semantic_layer_fetcher.get_metrics_compiled_sql(
        metrics=metrics,
        group_by=group_by,
        order_by=order_by,
        where=where,
        limit=limit,
    )
    if isinstance(result, GetMetricsCompiledSqlSuccess):
        return result.sql
    else:
        return result.error


SEMANTIC_LAYER_TOOLS = [
    list_metrics,
    list_saved_queries,
    get_dimensions,
    get_entities,
    query_metrics,
    get_metrics_compiled_sql,
]


def register_sl_tools(
    dbt_mcp: FastMCP,
    config_provider: ConfigProvider[SemanticLayerConfig],
    client_provider: SemanticLayerClientProvider,
    *,
    disabled_tools: set[ToolName],
    enabled_tools: set[ToolName],
    enabled_toolsets: set[Toolset],
    disabled_toolsets: set[Toolset],
) -> None:
    def bind_context() -> SemanticLayerToolContext:
        return SemanticLayerToolContext(
            config_provider=config_provider, client_provider=client_provider
        )

    register_tools(
        dbt_mcp,
        [tool.adapt_context(bind_context) for tool in SEMANTIC_LAYER_TOOLS],
        disabled_tools=disabled_tools,
        enabled_tools=enabled_tools,
        enabled_toolsets=enabled_toolsets,
        disabled_toolsets=disabled_toolsets,
    )



import json
import os
import subprocess
from collections.abc import Iterable
from typing import Any, Literal

from mcp.server.fastmcp import FastMCP
from pydantic import Field

from dbt_mcp.config.config import DbtCliConfig
from dbt_mcp.dbt_cli.binary_type import get_color_disable_flag
from dbt_mcp.dbt_cli.models.lineage_types import ModelLineage
from dbt_mcp.dbt_cli.models.manifest import Manifest
from dbt_mcp.prompts.prompts import get_prompt
from dbt_mcp.tools.annotations import create_tool_annotations
from dbt_mcp.tools.definitions import ToolDefinition
from dbt_mcp.tools.register import register_tools
from dbt_mcp.tools.tool_names import ToolName
from dbt_mcp.tools.toolsets import Toolset


def create_dbt_cli_tool_definitions(config: DbtCliConfig) -> list[ToolDefinition]:
    def _run_dbt_command(
        command: list[str],
        selector: str | None = None,
        resource_type: list[str] | None = None,
        is_selectable: bool = False,
        is_full_refresh: bool | None = False,
        vars: str | None = None,
    ) -> str:
        try:
            # Commands that should always be quiet to reduce output verbosity
            verbose_commands = [
                "build",
                "compile",
                "docs",
                "parse",
                "run",
                "test",
                "list",
            ]

            if is_full_refresh is True:
                command.append("--full-refresh")

            if vars and isinstance(vars, str):
                command.extend(["--vars", vars])

            if selector:
                selector_params = str(selector).split(" ")
                command.extend(["--select"] + selector_params)

            if isinstance(resource_type, Iterable):
                command.extend(["--resource-type"] + resource_type)

            full_command = command.copy()
            # Add --quiet flag to specific commands to reduce context window usage
            if len(full_command) > 0 and full_command[0] in verbose_commands:
                main_command = full_command[0]
                command_args = full_command[1:] if len(full_command) > 1 else []
                full_command = [main_command, "--quiet", *command_args]

            # We change the path only if this is an absolute path, otherwise we can have
            # problems with relative paths applied multiple times as DBT_PROJECT_DIR
            # is applied to dbt Core and Fusion as well (but not the dbt Cloud CLI)
            cwd_path = config.project_dir if os.path.isabs(config.project_dir) else None

            # Add appropriate color disable flag based on binary type
            color_flag = get_color_disable_flag(config.binary_type)
            args = [config.dbt_path, color_flag, *full_command]

            process = subprocess.Popen(
                args=args,
                cwd=cwd_path,
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                stdin=subprocess.DEVNULL,
                text=True,
            )
            output, _ = process.communicate(timeout=config.dbt_cli_timeout)
            return output or "OK"
        except subprocess.TimeoutExpired:
            return "Timeout: dbt command took too long to complete." + (
                " Try using a specific selector to narrow down the results."
                if is_selectable
                else ""
            )

    def build(
        selector: str | None = Field(
            default=None, description=get_prompt("dbt_cli/args/selectors")
        ),
        is_full_refresh: bool | None = Field(
            default=None, description=get_prompt("dbt_cli/args/full_refresh")
        ),
        vars: str | None = Field(
            default=None, description=get_prompt("dbt_cli/args/vars")
        ),
    ) -> str:
        return _run_dbt_command(
            ["build"],
            selector,
            is_selectable=True,
            is_full_refresh=is_full_refresh,
            vars=vars,
        )

    def compile() -> str:
        return _run_dbt_command(["compile"])

    def docs() -> str:
        return _run_dbt_command(["docs", "generate"])

    def ls(
        selector: str | None = Field(
            default=None, description=get_prompt("dbt_cli/args/selectors")
        ),
        resource_type: list[str] | None = Field(
            default=None,
            description=get_prompt("dbt_cli/args/resource_type"),
        ),
    ) -> str:
        return _run_dbt_command(
            ["list"],
            selector,
            resource_type=resource_type,
            is_selectable=True,
        )

    def parse() -> str:
        return _run_dbt_command(["parse"])

    def run(
        selector: str | None = Field(
            default=None, description=get_prompt("dbt_cli/args/selectors")
        ),
        is_full_refresh: bool | None = Field(
            default=None, description=get_prompt("dbt_cli/args/full_refresh")
        ),
        vars: str | None = Field(
            default=None, description=get_prompt("dbt_cli/args/vars")
        ),
    ) -> str:
        return _run_dbt_command(
            ["run"],
            selector,
            is_selectable=True,
            is_full_refresh=is_full_refresh,
            vars=vars,
        )

    def test(
        selector: str | None = Field(
            default=None, description=get_prompt("dbt_cli/args/selectors")
        ),
        vars: str | None = Field(
            default=None, description=get_prompt("dbt_cli/args/vars")
        ),
    ) -> str:
        return _run_dbt_command(["test"], selector, is_selectable=True, vars=vars)

    def show(
        sql_query: str = Field(description=get_prompt("dbt_cli/args/sql_query")),
        limit: int = Field(default=5, description=get_prompt("dbt_cli/args/limit")),
    ) -> str:
        args = ["show", "--inline", sql_query, "--favor-state"]
        # This is quite crude, but it should be okay for now
        # until we have a dbt Fusion integration.
        cli_limit = None
        if "limit" in sql_query.lower():
            # When --limit=-1, dbt won't apply a separate limit.
            cli_limit = -1
        elif limit:
            # This can be problematic if the LLM provides
            # a SQL limit and a `limit` argument. However, preferencing the limit
            # in the SQL query leads to a better experience when the LLM
            # makes that mistake.
            cli_limit = limit
        if cli_limit is not None:
            args.extend(["--limit", str(cli_limit)])
        args.extend(["--output", "json"])
        return _run_dbt_command(args)

    def _get_manifest() -> Manifest:
        """Helper function to load the dbt manifest.json file."""
        _run_dbt_command(["parse"])  # Ensure manifest is generated
        cwd_path = config.project_dir if os.path.isabs(config.project_dir) else None
        manifest_path = os.path.join(cwd_path or ".", "target", "manifest.json")
        with open(manifest_path) as f:
            manifest_data = json.load(f)
        return Manifest(**manifest_data)

    def get_model_lineage_dev(
        model_id: str,
        direction: Literal["parents", "children", "both"] = "both",
        exclude_prefixes: tuple[str, ...] = ("test.", "unit_test."),
        *,
        recursive: bool,
    ) -> dict[str, Any]:
        manifest = _get_manifest()
        model_lineage = ModelLineage.from_manifest(
            manifest,
            model_id,
            direction=direction,
            exclude_prefixes=exclude_prefixes,
            recursive=recursive,
        )
        return model_lineage.model_dump()

    def get_node_details_dev(
        node_id: str = Field(
            description=get_prompt("dbt_cli/args/node_id"),
        ),
    ) -> dict[str, Any]:
        # Comprehensive list of output keys to include all available node metadata
        output_keys = [
            # This is the entire list. Keeping it for ref and commmenting fields that are not needed.
            # Because we get the path, we should not need the raw_code and compiled_code that can be very big and that the LLM can read
            "unique_id",
            "name",
            "resource_type",
            "package_name",
            "original_file_path",
            "path",
            "alias",
            "description",
            "columns",
            "meta",
            "tags",
            "config",
            "depends_on",
            "patch_path",
            "schema",
            "database",
            "relation_name",
            # "raw_code",
            # "compiled_code",
            # "checksum",
            "language",
            "docs",
            "group",
            "access",
            "version",
            "latest_version",
            "deprecation_date",
            "contract",
            "constraints",
            "primary_key",
            "fqn",
            "build_path",
            "refs",
            "sources",
            "metrics",
            "created_at",
            "unrendered_config",
        ]
        output = _run_dbt_command(
            ["list", "--output", "json", "--output-keys", *output_keys],
            selector=node_id,
            is_selectable=True,
        )

        node_result: dict[str, Any] | None = None
        test_ids: list[str] = []

        for line in output.strip().split("\n"):
            if not line.strip():
                continue
            try:
                node = json.loads(line)
                resource_type = node.get("resource_type")
                if node_result is None:
                    # First node is the primary result
                    node_result = node
                elif resource_type == "test":
                    # Collect additional tests (associated with the primary node)
                    fqn = node.get("fqn", [])
                    test_ids.append(".".join(fqn) if fqn else node.get("unique_id", ""))
            except json.JSONDecodeError:
                continue

        if node_result is None:
            raise ValueError(f"No node found for selector: {node_id}")

        # Add test unique_ids for models, seeds, sources, and snapshots
        if node_result.get("resource_type") in ("model", "seed", "source", "snapshot"):
            node_result["tests"] = test_ids

        return node_result

    return [
        ToolDefinition(
            fn=build,
            description=get_prompt("dbt_cli/build"),
            annotations=create_tool_annotations(
                title="dbt build",
                read_only_hint=False,
                destructive_hint=True,
                idempotent_hint=False,
            ),
        ),
        ToolDefinition(
            fn=compile,
            description=get_prompt("dbt_cli/compile"),
            annotations=create_tool_annotations(
                title="dbt compile",
                read_only_hint=True,
                destructive_hint=False,
                idempotent_hint=True,
            ),
        ),
        ToolDefinition(
            fn=docs,
            description=get_prompt("dbt_cli/docs"),
            annotations=create_tool_annotations(
                title="dbt docs",
                read_only_hint=True,
                destructive_hint=False,
                idempotent_hint=True,
            ),
        ),
        ToolDefinition(
            name="list",
            fn=ls,
            description=get_prompt("dbt_cli/list"),
            annotations=create_tool_annotations(
                title="dbt list",
                read_only_hint=True,
                destructive_hint=False,
                idempotent_hint=True,
            ),
        ),
        ToolDefinition(
            fn=parse,
            description=get_prompt("dbt_cli/parse"),
            annotations=create_tool_annotations(
                title="dbt parse",
                read_only_hint=True,
                destructive_hint=False,
                idempotent_hint=True,
            ),
        ),
        ToolDefinition(
            fn=run,
            description=get_prompt("dbt_cli/run"),
            annotations=create_tool_annotations(
                title="dbt run",
                read_only_hint=False,
                destructive_hint=True,
                idempotent_hint=False,
            ),
        ),
        ToolDefinition(
            fn=test,
            description=get_prompt("dbt_cli/test"),
            annotations=create_tool_annotations(
                title="dbt test",
                read_only_hint=False,
                destructive_hint=True,
                idempotent_hint=False,
            ),
        ),
        ToolDefinition(
            fn=show,
            description=get_prompt("dbt_cli/show"),
            annotations=create_tool_annotations(
                title="dbt show",
                read_only_hint=True,
                destructive_hint=False,
                idempotent_hint=True,
            ),
        ),
        ToolDefinition(
            name="get_model_lineage_dev",
            fn=get_model_lineage_dev,
            description=get_prompt("dbt_cli/get_model_lineage_dev"),
            annotations=create_tool_annotations(
                title="Get Model Lineage (Dev)",
                read_only_hint=True,
                destructive_hint=False,
                idempotent_hint=True,
            ),
        ),
        ToolDefinition(
            name="get_node_details_dev",
            fn=get_node_details_dev,
            description=get_prompt("dbt_cli/get_node_details_dev"),
            annotations=create_tool_annotations(
                title="Get Node Details (Dev)",
                read_only_hint=True,
                destructive_hint=False,
                idempotent_hint=True,
            ),
        ),
    ]


def register_dbt_cli_tools(
    dbt_mcp: FastMCP,
    config: DbtCliConfig,
    *,
    disabled_tools: set[ToolName],
    enabled_tools: set[ToolName],
    enabled_toolsets: set[Toolset],
    disabled_toolsets: set[Toolset],
) -> None:
    register_tools(
        dbt_mcp,
        tool_definitions=create_dbt_cli_tool_definitions(config),
        disabled_tools=disabled_tools,
        enabled_tools=enabled_tools,
        enabled_toolsets=enabled_toolsets,
        disabled_toolsets=disabled_toolsets,
    )



import json
import os
import subprocess
from typing import Any

from mcp.server.fastmcp import FastMCP
from pydantic import Field

from dbt_mcp.config.config import DbtCodegenConfig
from dbt_mcp.dbt_cli.binary_type import get_color_disable_flag
from dbt_mcp.prompts.prompts import get_prompt
from dbt_mcp.tools.annotations import create_tool_annotations
from dbt_mcp.tools.definitions import ToolDefinition
from dbt_mcp.tools.register import register_tools
from dbt_mcp.tools.tool_names import ToolName
from dbt_mcp.tools.toolsets import Toolset


def create_dbt_codegen_tool_definitions(
    config: DbtCodegenConfig,
) -> list[ToolDefinition]:
    def _run_codegen_operation(
        macro_name: str,
        args: dict[str, Any] | None = None,
    ) -> str:
        """Execute a dbt-codegen macro using dbt run-operation."""
        try:
            # Build the dbt run-operation command
            command = ["run-operation", macro_name]

            # Add arguments if provided
            if args:
                # Convert args to JSON string for dbt
                args_json = json.dumps(args)
                command.extend(["--args", args_json])

            full_command = command.copy()
            # Add --quiet flag to reduce output verbosity
            main_command = full_command[0]
            command_args = full_command[1:] if len(full_command) > 1 else []
            full_command = [main_command, "--quiet", *command_args]

            # We change the path only if this is an absolute path, otherwise we can have
            # problems with relative paths applied multiple times as DBT_PROJECT_DIR
            # is applied to dbt Core and Fusion as well (but not the dbt Cloud CLI)
            cwd_path = config.project_dir if os.path.isabs(config.project_dir) else None

            # Add appropriate color disable flag based on binary type
            color_flag = get_color_disable_flag(config.binary_type)
            args_list = [config.dbt_path, color_flag, *full_command]

            process = subprocess.Popen(
                args=args_list,
                cwd=cwd_path,
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                stdin=subprocess.DEVNULL,
                text=True,
            )
            output, _ = process.communicate(timeout=config.dbt_cli_timeout)

            # Return the output directly or handle errors
            if process.returncode != 0:
                if "dbt found" in output and "resource" in output:
                    return f"Error: dbt-codegen package may not be installed. Run 'dbt deps' to install it.\n{output}"
                return f"Error running dbt-codegen macro: {output}"

            return output or "OK"

        except subprocess.TimeoutExpired:
            return f"Timeout: dbt-codegen operation took longer than {config.dbt_cli_timeout} seconds."
        except Exception as e:
            return str(e)

    def generate_source(
        schema_name: str = Field(
            description=get_prompt("dbt_codegen/args/schema_name")
        ),
        database_name: str | None = Field(
            default=None, description=get_prompt("dbt_codegen/args/database_name")
        ),
        table_names: list[str] | None = Field(
            default=None, description=get_prompt("dbt_codegen/args/table_names")
        ),
        generate_columns: bool = Field(
            default=False, description=get_prompt("dbt_codegen/args/generate_columns")
        ),
        include_descriptions: bool = Field(
            default=False,
            description=get_prompt("dbt_codegen/args/include_descriptions"),
        ),
    ) -> str:
        args: dict[str, Any] = {"schema_name": schema_name}
        if database_name:
            args["database_name"] = database_name
        if table_names:
            args["table_names"] = table_names
        args["generate_columns"] = generate_columns
        args["include_descriptions"] = include_descriptions

        return _run_codegen_operation("generate_source", args)

    def generate_model_yaml(
        model_names: list[str] = Field(
            description=get_prompt("dbt_codegen/args/model_names")
        ),
        upstream_descriptions: bool = Field(
            default=False,
            description=get_prompt("dbt_codegen/args/upstream_descriptions"),
        ),
        include_data_types: bool = Field(
            default=True, description=get_prompt("dbt_codegen/args/include_data_types")
        ),
    ) -> str:
        args: dict[str, Any] = {
            "model_names": model_names,
            "upstream_descriptions": upstream_descriptions,
            "include_data_types": include_data_types,
        }

        return _run_codegen_operation("generate_model_yaml", args)

    def generate_staging_model(
        source_name: str = Field(
            description=get_prompt("dbt_codegen/args/source_name")
        ),
        table_name: str = Field(description=get_prompt("dbt_codegen/args/table_name")),
        leading_commas: bool = Field(
            default=False, description=get_prompt("dbt_codegen/args/leading_commas")
        ),
        case_sensitive_cols: bool = Field(
            default=False,
            description=get_prompt("dbt_codegen/args/case_sensitive_cols"),
        ),
        materialized: str | None = Field(
            default=None, description=get_prompt("dbt_codegen/args/materialized")
        ),
    ) -> str:
        args: dict[str, Any] = {
            "source_name": source_name,
            "table_name": table_name,
            "leading_commas": leading_commas,
            "case_sensitive_cols": case_sensitive_cols,
        }
        if materialized:
            args["materialized"] = materialized

        return _run_codegen_operation("generate_base_model", args)

    return [
        ToolDefinition(
            fn=generate_source,
            description=get_prompt("dbt_codegen/generate_source"),
            annotations=create_tool_annotations(
                title="dbt-codegen generate_source",
                read_only_hint=True,
                destructive_hint=False,
                idempotent_hint=True,
            ),
        ),
        ToolDefinition(
            fn=generate_model_yaml,
            description=get_prompt("dbt_codegen/generate_model_yaml"),
            annotations=create_tool_annotations(
                title="dbt-codegen generate_model_yaml",
                read_only_hint=True,
                destructive_hint=False,
                idempotent_hint=True,
            ),
        ),
        ToolDefinition(
            fn=generate_staging_model,
            description=get_prompt("dbt_codegen/generate_staging_model"),
            annotations=create_tool_annotations(
                title="dbt-codegen generate_staging_model",
                read_only_hint=True,
                destructive_hint=False,
                idempotent_hint=True,
            ),
        ),
    ]


def register_dbt_codegen_tools(
    dbt_mcp: FastMCP,
    config: DbtCodegenConfig,
    *,
    disabled_tools: set[ToolName],
    enabled_tools: set[ToolName],
    enabled_toolsets: set[Toolset],
    disabled_toolsets: set[Toolset],
) -> None:
    register_tools(
        dbt_mcp,
        tool_definitions=create_dbt_codegen_tool_definitions(config),
        disabled_tools=disabled_tools,
        enabled_tools=enabled_tools,
        enabled_toolsets=enabled_toolsets,
        disabled_toolsets=disabled_toolsets,
    )



import logging


def disable_server_logs() -> None:
    # Disable uvicorn, fastapi, and related loggers
    loggers = (
        "uvicorn",
        "uvicorn.error",
        "uvicorn.access",
        "fastapi",
    )

    for logger_name in loggers:
        logging.getLogger(logger_name).disabled = True



import asyncio
import os

from dbt_mcp.config.config import load_config
from dbt_mcp.config.transport import validate_transport
from dbt_mcp.mcp.server import create_dbt_mcp


def main() -> None:
    config = load_config()
    server = asyncio.run(create_dbt_mcp(config))
    transport = validate_transport(os.environ.get("MCP_TRANSPORT", "stdio"))
    server.run(transport=transport)


if __name__ == "__main__":
    main()



import httpx

from dbt_mcp.config.config_providers import SemanticLayerConfig
from dbt_mcp.gql.errors import raise_gql_error


async def submit_request(
    sl_config: SemanticLayerConfig,
    payload: dict,
) -> dict:
    if "variables" not in payload:
        payload["variables"] = {}
    payload["variables"]["environmentId"] = sl_config.prod_environment_id

    async with httpx.AsyncClient() as client:
        response = await client.post(
            sl_config.url,
            json=payload,
            headers=sl_config.headers_provider.get_headers(),
        )
        response.raise_for_status()
        result = response.json()
        raise_gql_error(result)
        return result



import logging
from typing import Literal

logger = logging.getLogger(__name__)

TransportType = Literal["stdio", "sse", "streamable-http"]
VALID_TRANSPORTS: set[TransportType] = {"stdio", "sse", "streamable-http"}


def validate_transport(transport: str) -> TransportType:
    """Validate and return the MCP transport type."""
    transport = transport.strip().lower()

    if transport not in VALID_TRANSPORTS:
        valid_options = ", ".join(sorted(VALID_TRANSPORTS))
        raise ValueError(
            f"Invalid MCP_TRANSPORT: '{transport}'. Must be one of: {valid_options}"
        )

    logger.debug(f"Using MCP transport: {transport}")
    return transport  # type: ignore[return-value]



import asyncio
import base64
import json
from collections.abc import Callable
from contextlib import AbstractContextManager
from datetime import date, datetime, time, timedelta
from decimal import Decimal
from typing import Any, Protocol

import pyarrow as pa
from dbtsl.api.shared.query_params import (
    GroupByParam,
    OrderByGroupBy,
    OrderByMetric,
    OrderBySpec,
)
from dbtsl.client.sync import SyncSemanticLayerClient
from dbtsl.error import QueryFailedError

from dbt_mcp.config.config_providers import ConfigProvider, SemanticLayerConfig
from dbt_mcp.errors import InvalidParameterError
from dbt_mcp.semantic_layer.gql.gql import GRAPHQL_QUERIES
from dbt_mcp.semantic_layer.gql.gql_request import submit_request
from dbt_mcp.semantic_layer.levenshtein import get_misspellings
from dbt_mcp.semantic_layer.types import (
    DimensionToolResponse,
    EntityToolResponse,
    GetMetricsCompiledSqlError,
    GetMetricsCompiledSqlResult,
    GetMetricsCompiledSqlSuccess,
    MetricToolResponse,
    OrderByParam,
    QueryMetricsError,
    QueryMetricsResult,
    QueryMetricsSuccess,
    SavedQueryToolResponse,
)


def DEFAULT_RESULT_FORMATTER(table: pa.Table) -> str:
    """Convert PyArrow Table to JSON string with ISO date formatting.

    This replaces the pandas-based implementation with native PyArrow and Python json.
    Output format: array of objects (records), 2-space indentation, ISO date strings.
    """
    # Convert PyArrow table to list of dictionaries
    records = table.to_pylist()

    # Custom JSON encoder to handle date/datetime, time, Decimal, timedelta, and bytes objects
    class ExtendedJSONEncoder(json.JSONEncoder):
        def default(self, obj):
            if isinstance(obj, datetime | date):
                return obj.isoformat()
            if isinstance(obj, time):
                return obj.isoformat()
            if isinstance(obj, Decimal):
                return float(obj)
            if isinstance(obj, timedelta):
                return obj.total_seconds()
            if isinstance(obj, bytes):
                return base64.b64encode(obj).decode("utf-8")
            return super().default(obj)

    # Return JSON with records format and proper indentation
    return json.dumps(records, indent=2, cls=ExtendedJSONEncoder)


class SemanticLayerClientProtocol(Protocol):
    def session(self) -> AbstractContextManager[Any]: ...

    def query(
        self,
        metrics: list[str],
        group_by: list[GroupByParam | str] | None = None,
        limit: int | None = None,
        order_by: list[str | OrderByGroupBy | OrderByMetric] | None = None,
        where: list[str] | None = None,
        read_cache: bool = True,
    ) -> pa.Table: ...

    def compile_sql(
        self,
        metrics: list[str],
        group_by: list[str] | None = None,
        limit: int | None = None,
        order_by: list[str | OrderByGroupBy | OrderByMetric] | None = None,
        where: list[str] | None = None,
        read_cache: bool = True,
    ) -> str: ...


class SemanticLayerClientProvider(Protocol):
    async def get_client(self) -> SemanticLayerClientProtocol: ...


class DefaultSemanticLayerClientProvider:
    def __init__(self, config_provider: ConfigProvider[SemanticLayerConfig]):
        self.config_provider = config_provider

    async def get_client(self) -> SemanticLayerClientProtocol:
        config = await self.config_provider.get_config()
        return SyncSemanticLayerClient(
            environment_id=config.prod_environment_id,
            auth_token=config.token,
            host=config.host,
        )


class SemanticLayerFetcher:
    def __init__(
        self,
        config_provider: ConfigProvider[SemanticLayerConfig],
        client_provider: SemanticLayerClientProvider,
    ):
        self.client_provider = client_provider
        self.config_provider = config_provider
        self.entities_cache: dict[str, list[EntityToolResponse]] = {}
        self.dimensions_cache: dict[str, list[DimensionToolResponse]] = {}

    async def list_metrics(self, search: str | None = None) -> list[MetricToolResponse]:
        metrics_result = await submit_request(
            await self.config_provider.get_config(),
            {"query": GRAPHQL_QUERIES["metrics"], "variables": {"search": search}},
        )
        return [
            MetricToolResponse(
                name=m.get("name"),
                type=m.get("type"),
                label=m.get("label"),
                description=m.get("description"),
                metadata=(m.get("config") or {}).get("meta"),
            )
            for m in metrics_result["data"]["metricsPaginated"]["items"]
        ]

    async def list_saved_queries(
        self, search: str | None = None
    ) -> list[SavedQueryToolResponse]:
        """Fetch all saved queries from the Semantic Layer API."""
        saved_queries_result = await submit_request(
            await self.config_provider.get_config(),
            {
                "query": GRAPHQL_QUERIES["saved_queries"],
                "variables": {"search": search},
            },
        )
        return [
            SavedQueryToolResponse(
                name=sq.get("name"),
                label=sq.get("label"),
                description=sq.get("description"),
                metrics=[
                    m.get("name") for m in sq.get("queryParams", {}).get("metrics", [])
                ]
                if sq.get("queryParams", {}).get("metrics")
                else None,
                group_by=[
                    g.get("name") for g in sq.get("queryParams", {}).get("groupBy", [])
                ]
                if sq.get("queryParams", {}).get("groupBy")
                else None,
                where=sq.get("queryParams", {}).get("where", {}).get("whereSqlTemplate")
                if sq.get("queryParams", {}).get("where")
                else None,
            )
            for sq in saved_queries_result["data"]["savedQueriesPaginated"]["items"]
        ]

    async def get_dimensions(
        self, metrics: list[str], search: str | None = None
    ) -> list[DimensionToolResponse]:
        metrics_key = ",".join(sorted(metrics))
        if metrics_key not in self.dimensions_cache:
            dimensions_result = await submit_request(
                await self.config_provider.get_config(),
                {
                    "query": GRAPHQL_QUERIES["dimensions"],
                    "variables": {
                        "metrics": [{"name": m} for m in metrics],
                        "search": search,
                    },
                },
            )
            dimensions = []
            for d in dimensions_result["data"]["dimensionsPaginated"]["items"]:
                dimensions.append(
                    DimensionToolResponse(
                        name=d.get("name"),
                        type=d.get("type"),
                        description=d.get("description"),
                        label=d.get("label"),
                        granularities=d.get("queryableGranularities")
                        + d.get("queryableTimeGranularities"),
                        metadata=(d.get("config") or {}).get("meta"),
                    )
                )
            self.dimensions_cache[metrics_key] = dimensions
        return self.dimensions_cache[metrics_key]

    async def get_entities(
        self, metrics: list[str], search: str | None = None
    ) -> list[EntityToolResponse]:
        metrics_key = ",".join(sorted(metrics))
        if metrics_key not in self.entities_cache:
            entities_result = await submit_request(
                await self.config_provider.get_config(),
                {
                    "query": GRAPHQL_QUERIES["entities"],
                    "variables": {
                        "metrics": [{"name": m} for m in metrics],
                        "search": search,
                    },
                },
            )
            entities = [
                EntityToolResponse(
                    name=e.get("name"),
                    type=e.get("type"),
                    description=e.get("description"),
                )
                for e in entities_result["data"]["entitiesPaginated"]["items"]
            ]
            self.entities_cache[metrics_key] = entities
        return self.entities_cache[metrics_key]

    def _format_semantic_layer_error(self, error: Exception) -> str:
        """Format semantic layer errors by cleaning up common error message patterns."""
        error_str = str(error)
        return (
            error_str.replace("QueryFailedError(", "")
            .rstrip(")")
            .lstrip("[")
            .rstrip("]")
            .lstrip('"')
            .rstrip('"')
            .replace("INVALID_ARGUMENT: [FlightSQL]", "")
            .replace("(InvalidArgument; Prepare)", "")
            .replace("(InvalidArgument; ExecuteQuery)", "")
            .replace("Failed to prepare statement:", "")
            .replace("com.dbt.semanticlayer.exceptions.DataPlatformException:", "")
            .strip()
        )

    def _format_get_metrics_compiled_sql_error(
        self, compile_error: Exception
    ) -> GetMetricsCompiledSqlError:
        """Format get compiled SQL errors using the shared error formatter."""
        return GetMetricsCompiledSqlError(
            error=self._format_semantic_layer_error(compile_error)
        )

    async def validate_query_metrics_params(
        self, metrics: list[str], group_by: list[GroupByParam] | None
    ) -> str | None:
        errors = []
        available_metrics_names = [m.name for m in await self.list_metrics()]
        metric_misspellings = get_misspellings(
            targets=metrics,
            words=available_metrics_names,
            top_k=5,
        )
        for metric_misspelling in metric_misspellings:
            recommendations = (
                " Did you mean: " + ", ".join(metric_misspelling.similar_words) + "?"
            )
            errors.append(
                f"Metric {metric_misspelling.word} not found."
                + (recommendations if metric_misspelling.similar_words else "")
            )

        if errors:
            return f"Errors: {', '.join(errors)}"

        available_group_by = [d.name for d in await self.get_dimensions(metrics)] + [
            e.name for e in await self.get_entities(metrics)
        ]
        group_by_misspellings = get_misspellings(
            targets=[g.name for g in group_by or []],
            words=available_group_by,
            top_k=5,
        )
        for group_by_misspelling in group_by_misspellings:
            recommendations = (
                " Did you mean: " + ", ".join(group_by_misspelling.similar_words) + "?"
            )
            errors.append(
                f"Group by {group_by_misspelling.word} not found."
                + (recommendations if group_by_misspelling.similar_words else "")
            )

        if errors:
            return f"Errors: {', '.join(errors)}"
        return None

    # TODO: move this to the SDK
    def _format_query_failed_error(self, query_error: Exception) -> QueryMetricsError:
        if isinstance(query_error, QueryFailedError):
            return QueryMetricsError(
                error=self._format_semantic_layer_error(query_error)
            )
        else:
            return QueryMetricsError(error=str(query_error))

    def _get_order_bys(
        self,
        order_by: list[OrderByParam] | None,
        metrics: list[str] = [],
        group_by: list[GroupByParam] | None = None,
    ) -> list[OrderBySpec]:
        result: list[OrderBySpec] = []
        if order_by is None:
            return result
        queried_group_by = {g.name: g for g in group_by} if group_by else {}
        queried_metrics = set(metrics)
        for o in order_by:
            if o.name in queried_metrics:
                result.append(OrderByMetric(name=o.name, descending=o.descending))
            elif o.name in queried_group_by:
                selected_group_by = queried_group_by[o.name]
                result.append(
                    OrderByGroupBy(
                        name=selected_group_by.name,
                        descending=o.descending,
                        grain=selected_group_by.grain,
                    )
                )
            else:
                raise InvalidParameterError(
                    f"Order by `{o.name}` not found in metrics or group by"
                )
        return result

    async def get_metrics_compiled_sql(
        self,
        metrics: list[str],
        group_by: list[GroupByParam] | None = None,
        order_by: list[OrderByParam] | None = None,
        where: str | None = None,
        limit: int | None = None,
    ) -> GetMetricsCompiledSqlResult:
        """
        Get compiled SQL for the given metrics and group by parameters using the SDK.

        Args:
            metrics: List of metric names to get compiled SQL for
            group_by: List of group by parameters (dimensions/entities with optional grain)
            order_by: List of order by parameters
            where: Optional SQL WHERE clause to filter results
            limit: Optional limit for number of results

        Returns:
            GetMetricsCompiledSqlResult with either the compiled SQL or an error
        """
        validation_error = await self.validate_query_metrics_params(
            metrics=metrics,
            group_by=group_by,
        )
        if validation_error:
            return GetMetricsCompiledSqlError(error=validation_error)

        try:
            sl_client = await self.client_provider.get_client()
            with sl_client.session():
                parsed_order_by: list[OrderBySpec] = self._get_order_bys(
                    order_by=order_by, metrics=metrics, group_by=group_by
                )
                compiled_sql = await asyncio.to_thread(
                    sl_client.compile_sql,
                    metrics=metrics,
                    group_by=group_by,  # type: ignore
                    order_by=parsed_order_by,  # type: ignore
                    where=[where] if where else None,
                    limit=limit,
                    read_cache=True,
                )

                return GetMetricsCompiledSqlSuccess(sql=compiled_sql)

        except Exception as e:
            return self._format_get_metrics_compiled_sql_error(e)

    async def query_metrics(
        self,
        metrics: list[str],
        group_by: list[GroupByParam] | None = None,
        order_by: list[OrderByParam] | None = None,
        where: str | None = None,
        limit: int | None = None,
        result_formatter: Callable[[pa.Table], str] | None = None,
    ) -> QueryMetricsResult:
        validation_error = await self.validate_query_metrics_params(
            metrics=metrics,
            group_by=group_by,
        )
        if validation_error:
            return QueryMetricsError(error=validation_error)

        try:
            query_error = None
            sl_client = await self.client_provider.get_client()
            with sl_client.session():
                # Catching any exception within the session
                # to ensure it is closed properly
                try:
                    parsed_order_by: list[OrderBySpec] = self._get_order_bys(
                        order_by=order_by, metrics=metrics, group_by=group_by
                    )
                    query_result = await asyncio.to_thread(
                        sl_client.query,
                        metrics=metrics,
                        group_by=group_by,  # type: ignore
                        order_by=parsed_order_by,  # type: ignore
                        where=[where] if where else None,
                        limit=limit,
                    )
                except Exception as e:
                    query_error = e
            if query_error:
                return self._format_query_failed_error(query_error)
            formatter = result_formatter or DEFAULT_RESULT_FORMATTER
            json_result = formatter(query_result)
            return QueryMetricsSuccess(result=json_result or "")
        except Exception as e:
            return self._format_query_failed_error(e)



"""Discovery/Metadata API tool errors."""

from dbt_mcp.errors.base import ToolCallError


class DiscoveryToolCallError(ToolCallError):
    """Base exception for Discovery/Metadata API tool errors."""

    pass


class GraphQLError(DiscoveryToolCallError):
    """Exception raised for GraphQL API and query errors."""

    pass



import subprocess
from enum import Enum

from dbt_mcp.errors import BinaryExecutionError


class BinaryType(Enum):
    DBT_CORE = "dbt_core"
    FUSION = "fusion"
    DBT_CLOUD_CLI = "dbt_cloud_cli"


def detect_binary_type(file_path: str) -> BinaryType:
    """
    Detect the type of dbt binary (dbt Core, Fusion, or dbt Cloud CLI) by running --help.

    Args:
        file_path: Path to the dbt executable

    Returns:
        BinaryType: The detected binary type

    Raises:
        Exception: If the binary cannot be executed or accessed
    """
    try:
        result = subprocess.run(
            [file_path, "--help"],
            check=False,
            capture_output=True,
            text=True,
            timeout=60,
        )
        help_output = result.stdout
    except Exception as e:
        raise BinaryExecutionError(f"Cannot execute binary {file_path}: {e}")

    if not help_output:
        # Default to dbt Core if no output
        return BinaryType.DBT_CORE

    first_line = help_output.split("\n")[0] if help_output else ""

    # Check for dbt-fusion
    if "dbt-fusion" in first_line:
        return BinaryType.FUSION

    # Check for dbt Core
    if "Usage: dbt [OPTIONS] COMMAND [ARGS]..." in first_line:
        return BinaryType.DBT_CORE

    # Check for dbt Cloud CLI
    if "The dbt Cloud CLI" in first_line:
        return BinaryType.DBT_CLOUD_CLI

    # Default to dbt Core - We could move to Fusion in the future
    return BinaryType.DBT_CORE


def get_color_disable_flag(binary_type: BinaryType) -> str:
    """
    Get the appropriate color disable flag for the given binary type.

    Args:
        binary_type: The type of dbt binary

    Returns:
        str: The color disable flag to use
    """
    if binary_type == BinaryType.DBT_CLOUD_CLI:
        return "--no-color"
    else:  # DBT_CORE or FUSION
        return "--no-use-colors"



"""MCP Server tools for server information and management."""

import logging
from importlib.metadata import version

from mcp.server.fastmcp import FastMCP

from dbt_mcp.tools.definitions import dbt_mcp_tool
from dbt_mcp.tools.register import register_tools
from dbt_mcp.tools.tool_names import ToolName
from dbt_mcp.tools.toolsets import Toolset

logger = logging.getLogger(__name__)


def _get_server_version() -> str:
    """Get the dbt-mcp server version from package metadata."""
    try:
        return version("dbt-mcp")
    except Exception:
        return "unknown"


@dbt_mcp_tool(
    description="Get the version of the dbt MCP server. Use this to check what version of the dbt MCP server is running.",
    title="Get MCP Server Version",
    read_only_hint=True,
    destructive_hint=False,
    idempotent_hint=True,
    open_world_hint=False,
)
def get_mcp_server_version() -> str:
    """Returns the current version of the dbt MCP server."""
    return _get_server_version()


MCP_SERVER_TOOLS = [
    get_mcp_server_version,
]


def register_mcp_server_tools(
    dbt_mcp: FastMCP,
    *,
    disabled_tools: set[ToolName],
    enabled_tools: set[ToolName],
    enabled_toolsets: set[Toolset],
    disabled_toolsets: set[Toolset],
) -> None:
    """Register MCP server tools."""
    register_tools(
        dbt_mcp,
        tool_definitions=MCP_SERVER_TOOLS,
        disabled_tools=disabled_tools,
        enabled_tools=enabled_tools,
        enabled_toolsets=enabled_toolsets,
        disabled_toolsets=disabled_toolsets,
    )



"""Local LSP Connection Provider Implementation.

This module provides the concrete implementation of LSPConnectionProvider for
managing socket-based connections to a local LSP server process.
"""

import logging


from dbt_mcp.lsp.lsp_binary_manager import LspBinaryInfo
from dbt_mcp.lsp.lsp_connection import SocketLSPConnection
from dbt_mcp.lsp.providers.lsp_connection_provider import (
    LSPConnectionProvider,
    LSPConnectionProviderProtocol,
)

logger = logging.getLogger(__name__)


class LocalLSPConnectionProvider(LSPConnectionProvider):
    """Provider for managing a local LSP server connection.

    This provider implements a singleton pattern for LSP connections:
    - Only one connection is created and reused across all requests
    - Connection is lazily initialized on first use
    - Connection lifecycle is tied to the MCP server lifecycle

    The provider is owned by the DbtMCP server instance and is cleaned up
    during server shutdown, ensuring no orphaned LSP server processes.

    Attributes:
        lsp_connection: The singleton connection instance (None until first use)
        lsp_binary_info: Information about the LSP binary (path and version)
        project_dir: dbt project directory for the LSP server
    """

    def __init__(self, lsp_binary_info: LspBinaryInfo, project_dir: str):
        """Initialize the local LSP connection provider.

        Note: This does NOT create the connection - it only stores the
        configuration needed to create it later (lazy initialization).

        Args:
            lsp_binary_info: Information about the LSP binary to use
            project_dir: Path to the dbt project directory
        """
        self.lsp_connection: LSPConnectionProviderProtocol | None = None
        self.lsp_binary_info = lsp_binary_info
        self.project_dir = project_dir

    async def _new_connection(self) -> LSPConnectionProviderProtocol:
        """Create and initialize a new LSP connection.

        This is an internal method that handles the actual connection creation
        and initialization sequence:
        1. Create SocketLSPConnection with binary info
        2. Start the LSP server process and establish socket connection
        3. Send LSP initialize request and wait for server capabilities

        Returns:
            A fully initialized LSP connection

        Raises:
            RuntimeError: If connection creation or initialization fails
        """
        # Defensive check: This shouldn't happen due to get_connection() logic,
        # but included for clarity and safety
        if self.lsp_connection is not None:
            return self.lsp_connection

        try:
            logger.info(
                f"Using LSP binary in {self.lsp_binary_info.path} with version {self.lsp_binary_info.version}"
            )

            # Create the connection wrapper (doesn't start the process yet)
            lsp_connection = SocketLSPConnection(
                binary_path=self.lsp_binary_info.path,
                args=[],
                cwd=self.project_dir,
            )

            # Start the LSP server process and establish socket connection
            # This is when the actual subprocess is spawned
            await lsp_connection.start()
            logger.info("LSP connection started successfully")

            # Send the LSP initialize request to exchange capabilities
            # The server won't accept other requests until this completes
            await lsp_connection.initialize()
            logger.info("LSP connection initialized successfully")

            return lsp_connection
        except Exception as e:
            logger.error(f"Error starting LSP connection: {e}")
            # Clean up any partial state to ensure clean retry
            self.lsp_connection = None
            raise RuntimeError("Error: Failed to establish LSP connection")

    async def get_connection(self) -> LSPConnectionProviderProtocol:
        """Get the LSP connection, creating it if needed (lazy initialization).

        This implements the singleton pattern:
        - First call: Creates, starts, and initializes the connection
        - Subsequent calls: Returns the existing connection

        The connection is only created when actually needed (typically when
        a tool is first invoked), not during provider or server initialization.

        Returns:
            The singleton LSP connection instance

        Raises:
            RuntimeError: If connection creation fails
        """
        if self.lsp_connection is None:
            # Lazy initialization: Create the connection on first use
            # This happens AFTER the MCP server is listening for requests
            self.lsp_connection = await self._new_connection()
        return self.lsp_connection

    async def cleanup_connection(self) -> None:
        """Cleanup and stop the LSP connection if it exists.

        This method is called during MCP server shutdown via app_lifespan.
        It gracefully stops the LSP server process and cleans up resources.

        The method is idempotent and safe to call multiple times or when
        no connection was ever created (handles None case).

        Note: Exceptions during cleanup are caught and logged but not re-raised
        to ensure shutdown proceeds even if cleanup fails.
        """
        if self.lsp_connection:
            try:
                logger.info("Cleaning up LSP connection")
                # Stop the LSP server process and close socket
                await self.lsp_connection.stop()
            except Exception as e:
                # Log but don't re-raise - we want shutdown to continue
                logger.error(f"Error cleaning up LSP connection: {e}")
            finally:
                # Always clear the reference, even if stop() failed
                # This ensures we don't try to reuse a failed connection
                self.lsp_connection = None



import asyncio
import logging
import re
from typing import Any

from dbt_mcp.config.config_providers import AdminApiConfig
from dbt_mcp.dbt_admin.client import DbtAdminAPIClient

from dbt_mcp.dbt_admin.constants import (
    STATUS_MAP,
    TRUNCATED_LOGS_LENGTH,
    JobRunStatus,
    RunResultsStatus,
)
from dbt_mcp.dbt_admin.run_artifacts.config import (
    OutputResultSchema,
    OutputStepSchema,
    RunDetailsSchema,
    RunResultsArtifactSchema,
    RunResultSchema,
    RunStepSchema,
    SourcesArtifactSchema,
)
from dbt_mcp.errors import ArtifactRetrievalError
from pydantic import ValidationError

logger = logging.getLogger(__name__)


class JobRunFetcher:
    """Base class for fetching and parsing dbt Cloud job run artifacts.

    Provides shared functionality for ErrorFetcher and WarningFetcher to:
    - Fetch run_results.json artifacts from specific steps
    - Fetch sources.json artifacts for source freshness checks
    - Parse source freshness results (errors/warnings)

    Both ErrorFetcher and WarningFetcher inherit from this class but process
    different sets of steps:
    - ErrorFetcher: processes failed steps (status == ERROR)
    - WarningFetcher: processes successful steps (status == SUCCESS)

    Each fetcher instance is single-use and processes one job run analysis.
    """

    def __init__(
        self,
        run_id: int,
        run_details: dict[str, Any],
        client: DbtAdminAPIClient,
        admin_api_config: AdminApiConfig,
    ):
        """Initialize the job run fetcher.

        Args:
            run_id: The dbt Cloud job run ID
            run_details: Raw job run details from the dbt Admin API
            client: dbt Admin API client for making artifact requests
            admin_api_config: Configuration containing account_id and credentials
        """
        self.run_id = run_id
        self.run_details = run_details
        self.client = client
        self.admin_api_config = admin_api_config

    async def _fetch_run_results_artifact(self, step: RunStepSchema) -> str | None:
        """Fetch run_results.json artifact for the step."""
        step_index = step.index

        try:
            if step_index is not None:
                run_results_content = await self.client.get_job_run_artifact(
                    self.admin_api_config.account_id,
                    self.run_id,
                    "run_results.json",
                    step=step_index,
                )
                logger.info(f"Got run_results.json from step {step_index}")
                return run_results_content
            return None
        except ArtifactRetrievalError as e:
            logger.debug(f"No run_results.json for step {step_index}: {e}")
            return None

    async def _fetch_sources_artifact(self, step: RunStepSchema) -> str | None:
        """Fetch sources.json artifact for the step."""
        step_index = step.index

        try:
            if step_index is not None:
                sources_content = await self.client.get_job_run_artifact(
                    self.admin_api_config.account_id,
                    self.run_id,
                    "sources.json",
                    step=step_index,
                )
                logger.info(f"Got sources.json from step {step_index}")
                return sources_content
            return None
        except ArtifactRetrievalError as e:
            logger.debug(f"No sources.json for step {step_index}: {e}")
            return None

    async def _check_source_freshness_results(
        self,
        step: RunStepSchema,
        status_filter: list[str],
        message_prefix: str,
        result_status: str | None = None,
    ) -> OutputStepSchema | None:
        """
        Check for source freshness results (errors/warnings) in sources.json artifact.
        Note: Only method used by both ErrorFetcher and WarningFetcher that differ in usage
        with params so including a complete docstring.

        Args:
            step: The run step to check for source freshness results
            status_filter: List of status values to filter for (["error", "fail"] or ["warn"])
            message_prefix: Prefix for the result message ("Source freshness error")
            result_status: Optional status value to set in OutputResultSchema ("warn")
        """
        sources_content = await self._fetch_sources_artifact(step)

        if not sources_content:
            return None

        try:
            sources = SourcesArtifactSchema.model_validate_json(sources_content)

            message_suffix = (
                " (max allowed exceeded)" if "error" in message_prefix.lower() else ""
            )

            results = [
                OutputResultSchema(
                    unique_id=result.unique_id,
                    relation_name=result.unique_id.split(".")[-1]
                    if result.unique_id
                    else "Unknown",
                    message=f"{message_prefix}: {result.max_loaded_at_time_ago_in_s or 0:.0f}s since last load{message_suffix}",
                    status=result_status,
                )
                for result in sources.results
                if result.status in status_filter
            ]

            if results:
                return OutputStepSchema(
                    target=None,
                    step_name=step.name,
                    finished_at=step.finished_at,
                    results=results,
                )
            return None
        except ValidationError as e:
            logger.debug(f"sources.json validation failed: {e}")
            return None
        except Exception as e:
            logger.debug(f"Error parsing sources.json: {e}")
            return None


class ErrorFetcher(JobRunFetcher):
    """Parses dbt Cloud job run data to extract focused error information."""

    async def analyze_run_errors(self) -> dict[str, Any]:
        """Parse the run data and return all failed steps with their details."""
        try:
            run_details = RunDetailsSchema.model_validate(self.run_details)
            failed_steps = self._find_failed_steps(run_details)

            if run_details.is_cancelled:
                error_result = self._create_error_result(
                    message="Job run was cancelled",
                    finished_at=run_details.finished_at,
                )
                return {"failed_steps": [error_result.model_dump()]}

            if not failed_steps:
                error_result = self._create_error_result("No failed step found")
                return {"failed_steps": [error_result.model_dump()]}

            # asyncio.gather to process steps in parallel
            step_results = await asyncio.gather(
                *[self._get_step_errors(step) for step in failed_steps],
                return_exceptions=True,
            )

            processed_steps: list[OutputStepSchema] = []
            for step_result in step_results:
                if isinstance(step_result, BaseException):
                    logger.warning(f"Error processing failed step: {step_result}")
                    continue
                processed_steps.append(step_result)

            return {"failed_steps": [step.model_dump() for step in processed_steps]}

        except ValidationError as e:
            logger.error(f"Schema validation failed for run {self.run_id}: {e}")
            error_result = self._create_error_result(f"Validation failed: {e!s}")
            return {"failed_steps": [error_result.model_dump()]}
        except Exception as e:
            logger.error(f"Error analyzing run {self.run_id}: {e}")
            error_result = self._create_error_result(str(e))
            return {"failed_steps": [error_result.model_dump()]}

    def _find_failed_steps(self, run_details: RunDetailsSchema) -> list[RunStepSchema]:
        """Find all failed steps in the run."""
        failed_steps = []
        for step in run_details.run_steps:
            if step.status == STATUS_MAP[JobRunStatus.ERROR]:
                failed_steps.append(step)
        return failed_steps

    async def _get_step_errors(self, failed_step: RunStepSchema) -> OutputStepSchema:
        """Get simplified error information from failed step."""
        run_results_content = await self._fetch_run_results_artifact(failed_step)

        if not run_results_content:
            # Try to get structured errors from sources.json for source freshness checks
            source_errors = await self._check_source_freshness_errors(failed_step)
            if source_errors:
                return source_errors
            # Fall back to truncated logs if no structured errors available
            return self._handle_artifact_error(failed_step)

        return self._parse_run_results(run_results_content, failed_step)

    def _parse_run_results(
        self, run_results_content: str, failed_step: RunStepSchema
    ) -> OutputStepSchema:
        """Parse run_results.json content and extract errors."""
        try:
            run_results = RunResultsArtifactSchema.model_validate_json(
                run_results_content
            )
            errors = self._extract_errors_from_results(run_results.results)

            return self._build_error_response(errors, failed_step, run_results.args)

        except ValidationError as e:
            logger.warning(f"run_results.json validation failed: {e}")
            return self._handle_artifact_error(failed_step)
        except Exception as e:
            logger.warning(f"run_results.json parsing failed: {e}")
            return self._handle_artifact_error(failed_step)

    def _extract_errors_from_results(
        self, results: list[RunResultSchema]
    ) -> list[OutputResultSchema]:
        """Extract error results from run results."""
        errors = []
        for result in results:
            if result.status in [
                RunResultsStatus.ERROR.value,
                RunResultsStatus.FAIL.value,
            ]:
                relation_name = (
                    result.relation_name
                    if result.relation_name is not None
                    else "No database relation"
                )
                error = OutputResultSchema(
                    unique_id=result.unique_id,
                    relation_name=relation_name,
                    message=result.message or "",
                    compiled_code=result.compiled_code,
                )
                errors.append(error)
        return errors

    def _build_error_response(
        self,
        errors: list[OutputResultSchema],
        failed_step: RunStepSchema,
        args: Any | None,
    ) -> OutputStepSchema:
        """Build the final error response structure."""
        target = args.target if args else None
        step_name = failed_step.name
        finished_at = failed_step.finished_at
        truncated_logs = self._get_truncated_logs(failed_step)

        if errors:
            return OutputStepSchema(
                results=errors,
                step_name=step_name,
                finished_at=finished_at,
                target=target,
            )

        message = "No failures found in run_results.json"

        return self._create_error_result(
            message=message,
            target=target,
            step_name=step_name,
            finished_at=finished_at,
            truncated_logs=truncated_logs,
        )

    def _create_error_result(
        self,
        message: str,
        unique_id: str | None = None,
        relation_name: str | None = None,
        target: str | None = None,
        step_name: str | None = None,
        finished_at: str | None = None,
        compiled_code: str | None = None,
        truncated_logs: str | None = None,
    ) -> OutputStepSchema:
        """Create a standardized error results using OutputStepSchema."""
        error = OutputResultSchema(
            unique_id=unique_id,
            relation_name=relation_name,
            message=message,
            compiled_code=compiled_code,
            truncated_logs=truncated_logs,
        )
        return OutputStepSchema(
            results=[error],
            step_name=step_name,
            finished_at=finished_at,
            target=target,
        )

    async def _check_source_freshness_errors(
        self, step: RunStepSchema
    ) -> OutputStepSchema | None:
        """Check for source freshness errors in sources.json artifact."""
        return await self._check_source_freshness_results(
            step=step,
            status_filter=["error", "fail"],  # sources.json uses "fail" for errors
            message_prefix="Source freshness error",
            result_status=None,
        )

    def _handle_artifact_error(self, failed_step: RunStepSchema) -> OutputStepSchema:
        """Handle cases where run_results.json is not available."""
        relation_name = "No database relation"
        step_name = failed_step.name
        finished_at = failed_step.finished_at
        truncated_logs = self._get_truncated_logs(failed_step)

        message = "run_results.json not available - returning logs"

        return self._create_error_result(
            message=message,
            relation_name=relation_name,
            step_name=step_name,
            finished_at=finished_at,
            truncated_logs=truncated_logs,
        )

    def _get_truncated_logs(self, step: RunStepSchema) -> str | None:
        """Truncate logs to the last N lines."""
        split_logs = step.logs.splitlines() if step.logs else []
        if len(split_logs) > TRUNCATED_LOGS_LENGTH:  # N = 50
            split_logs = [
                f"Logs truncated to last {TRUNCATED_LOGS_LENGTH} lines..."
            ] + split_logs[-TRUNCATED_LOGS_LENGTH:]
        return "\n".join(split_logs)


class WarningFetcher(JobRunFetcher):
    """Parser for extracting warning information from successful runs."""

    async def analyze_run_warnings(self) -> dict[str, Any]:
        """Parse the run data and return all structured / log warnings with their details."""
        try:
            run_details = RunDetailsSchema.model_validate(self.run_details)

            if run_details.is_cancelled:
                return self._empty_response("Run was cancelled")

            successful_steps = self._find_successful_steps(run_details)
            if not successful_steps:
                return self._empty_response("No successful steps found")

            warning_step_models, all_log_warning_results = await self._collect_warnings(
                successful_steps
            )
            deduplicated_log_warnings = self._deduplicate_log_warnings(
                all_log_warning_results
            )
            summary = self._create_summary(
                warning_step_models, deduplicated_log_warnings
            )

            return {
                "has_warnings": summary["total_warnings"] > 0,
                "warning_steps": [step.model_dump() for step in warning_step_models],
                "log_warnings": [
                    warning.model_dump() for warning in deduplicated_log_warnings
                ],
                "summary": summary,
            }

        except ValidationError as e:
            logger.error(f"Schema validation failed for run {self.run_id}: {e}")
            return self._empty_response(f"Validation failed: {e!s}")
        except Exception as e:
            logger.error(f"Error analyzing warnings for run {self.run_id}: {e}")
            return self._empty_response(str(e))

    def _find_successful_steps(
        self, run_details: RunDetailsSchema
    ) -> list[RunStepSchema]:
        """Find all successful steps in the run."""
        return [
            step
            for step in run_details.run_steps
            if step.status == STATUS_MAP[JobRunStatus.SUCCESS]
        ]

    async def _collect_warnings(
        self, successful_steps: list[RunStepSchema]
    ) -> tuple[list[OutputStepSchema], list[OutputResultSchema]]:
        """Collect warnings from all successful steps."""
        # asyncio.gather to process steps in parallel
        step_results = await asyncio.gather(
            *[self._get_step_warnings(step) for step in successful_steps],
            return_exceptions=True,
        )

        warning_steps: list[OutputStepSchema] = []
        all_log_warnings: list[OutputResultSchema] = []

        for step_result in step_results:
            if isinstance(step_result, BaseException):
                logger.warning(f"Error processing step: {step_result}")
                continue

            if step_result is not None:
                structured_warnings = [
                    warning for warning in step_result.results if warning.unique_id
                ]
                step_log_warnings = [
                    warning for warning in step_result.results if not warning.unique_id
                ]

                if structured_warnings:
                    warning_steps.append(
                        OutputStepSchema(
                            target=step_result.target,
                            step_name=step_result.step_name,
                            finished_at=step_result.finished_at,
                            results=structured_warnings,
                        )
                    )

                if step_log_warnings:
                    all_log_warnings.extend(step_log_warnings)

        return warning_steps, all_log_warnings

    async def _get_step_warnings(self, step: RunStepSchema) -> OutputStepSchema | None:
        """Get warnings from a single successful step."""
        run_results_content = await self._fetch_run_results_artifact(step)

        result: OutputStepSchema | None = None
        if run_results_content:
            result = self._parse_run_results_for_warnings(run_results_content, step)
        else:
            result = await self._check_source_freshness_warnings(step)

        log_warnings = self._extract_log_warnings(step)

        if log_warnings:
            if result is not None:
                combined = result.results + log_warnings
                deduplicated = self._deduplicate_warning_results(combined)
                result = result.model_copy(update={"results": deduplicated})
            else:
                deduplicated = self._deduplicate_warning_results(log_warnings)
                result = OutputStepSchema(
                    target=None,
                    step_name=step.name,
                    finished_at=step.finished_at,
                    results=deduplicated,
                )
        elif result is not None:
            deduplicated = self._deduplicate_warning_results(result.results)
            result = result.model_copy(update={"results": deduplicated})

        return result

    def _parse_run_results_for_warnings(
        self, run_results_content: str, step: RunStepSchema
    ) -> OutputStepSchema | None:
        """Parse run_results.json content and extract structured warnings."""
        try:
            run_results = RunResultsArtifactSchema.model_validate_json(
                run_results_content
            )
            warnings = self._extract_warnings_from_results(run_results.results)

            if warnings:
                return OutputStepSchema(
                    target=run_results.args.target if run_results.args else None,
                    step_name=step.name,
                    finished_at=step.finished_at,
                    results=warnings,
                )
            return None
        except (ValidationError, Exception) as e:
            logger.warning(f"run_results.json parsing failed: {e}")
            return None

    def _extract_warnings_from_results(
        self, results: list[RunResultSchema]
    ) -> list[OutputResultSchema]:
        """Extract warning results from run results."""
        warnings = []
        for result in results:
            if result.status == RunResultsStatus.WARN.value:
                warnings.append(
                    OutputResultSchema(
                        unique_id=result.unique_id,
                        relation_name=result.relation_name or "No database relation",
                        message=result.message or "Warning detected",
                        status="warn",
                        compiled_code=result.compiled_code,
                    )
                )
        return warnings

    async def _check_source_freshness_warnings(
        self, step: RunStepSchema
    ) -> OutputStepSchema | None:
        """Check for source freshness warnings in sources.json artifact."""
        return await self._check_source_freshness_results(
            step=step,
            status_filter=["warn"],
            message_prefix="Source freshness warning",
            result_status="warn",
        )

    def _extract_log_warnings(self, step: RunStepSchema) -> list[OutputResultSchema]:
        """Extract warnings from step logs, matching on [WARNING] log entries."""
        if not step.logs:
            return []

        if "[WARNING]" not in step.logs:
            return []

        # TODO: Fusion logs differ from core (currently core only)
        # Need to explore and implement a solution for this

        # Remove ANSI color codes to focus on text
        ansi_escape = re.compile(r"\x1b\[[0-9;]*m")
        clean_logs = ansi_escape.sub("", step.logs)
        lines = clean_logs.split("\n")
        warnings = []
        warning_pattern = r"\[WARNING\]"  # Standard warning log pattern in dbt Platform

        i = 0
        while i < len(lines):
            line = lines[i]

            if re.search(warning_pattern, line, re.IGNORECASE):
                message_lines = [line]
                j = i + 1

                # Capture continuation lines until next log entry or 5-line limit
                while j < len(lines):
                    next_line = lines[j]
                    # Timestamp pattern (HH:MM:SS) marks start of new log entry
                    if (
                        re.match(r"^\d{2}:\d{2}:\d{2}\s", next_line)
                        or len(message_lines) >= 5
                    ):
                        break
                    message_lines.append(next_line)
                    j += 1

                warnings.append(
                    OutputResultSchema(
                        unique_id=None,
                        relation_name=None,
                        message="\n".join(message_lines).strip(),
                        status="warn",
                    )
                )

                i = j
            else:
                i += 1

        return warnings

    def _deduplicate_warning_results(
        self, warnings: list[OutputResultSchema]
    ) -> list[OutputResultSchema]:
        """Deduplicate warnings based on unique_id or message content from run_results.json."""
        seen_unique_ids = set()
        seen_messages = set()
        deduplicated = []

        for warning in warnings:
            unique_id = warning.unique_id
            message = warning.message or ""

            if unique_id:
                if unique_id not in seen_unique_ids:
                    seen_unique_ids.add(unique_id)
                    deduplicated.append(warning)
            else:
                normalized_message = " ".join(message.split()).lower()
                if normalized_message and normalized_message not in seen_messages:
                    seen_messages.add(normalized_message)
                    deduplicated.append(warning)

        return deduplicated

    def _deduplicate_log_warnings(
        self, log_warnings: list[OutputResultSchema]
    ) -> list[OutputResultSchema]:
        """Deduplicate log warnings based on first line of text message content."""

        # Trade-off w/ current approach:
        # There is a chance that two different warnings with same first line get deduplicated
        # For example, if two deprecation warnings have an identical first line
        # but differ later in the nodes they refer to
        # However, so far in testing, this has not been an issue and greatly reduces noise
        # between repeated warnings in logs in multi-step runs

        seen_warning_keys = set()
        deduplicated = []

        for warning in log_warnings:
            message = warning.message or ""
            # Use only first line as the "core" identifier for deduplication
            first_line = message.split("\n")[0] if message else ""
            # Strip timestamps so same warning at different times matches
            first_line_clean = re.sub(r"^\d{2}:\d{2}:\d{2}\s+", "", first_line)
            # Normalize whitespace and case for matching
            normalized_key = " ".join(first_line_clean.split()).lower()

            if normalized_key and normalized_key not in seen_warning_keys:
                seen_warning_keys.add(normalized_key)
                deduplicated.append(warning)

        return deduplicated

    def _create_summary(
        self,
        warning_steps: list[OutputStepSchema],
        log_warning_steps: list[OutputResultSchema],
    ) -> dict[str, int]:
        """Create a summary of warning counts."""
        test_warnings = 0
        freshness_warnings = 0

        for step_result in warning_steps:
            for warning in step_result.results:
                unique_id = warning.unique_id or ""
                if "test." in unique_id:
                    test_warnings += 1
                elif "source." in unique_id:
                    freshness_warnings += 1

        log_warnings = len(log_warning_steps)
        total_warnings = test_warnings + freshness_warnings + log_warnings

        return {
            "total_warnings": total_warnings,
            "test_warnings": test_warnings,
            "freshness_warnings": freshness_warnings,
            "log_warnings": log_warnings,
        }

    def _empty_response(self, reason: str = "No warnings found") -> dict[str, Any]:
        """Return an empty warning response with a reason."""
        return {
            "has_warnings": False,
            "warning_steps": [],
            "log_warnings": [],
            "summary": {
                "total_warnings": 0,
                "test_warnings": 0,
                "freshness_warnings": 0,
                "log_warnings": 0,
            },
            "message": reason,
        }



import logging
from collections.abc import Sequence
from enum import Enum

from mcp.server.fastmcp import FastMCP

from dbt_mcp.tools.definitions import GenericToolDefinition
from dbt_mcp.tools.tool_names import ToolName
from dbt_mcp.tools.toolsets import TOOL_TO_TOOLSET, Toolset

logger = logging.getLogger(__name__)


def should_register_tool[NameEnum: Enum](
    tool_name: NameEnum,
    enabled_tools: set[NameEnum],
    disabled_tools: set[NameEnum],
    enabled_toolsets: set[Toolset],
    disabled_toolsets: set[Toolset],
    tool_to_toolset: dict[NameEnum, Toolset],
) -> bool:
    """Determine if a tool should be registered based on precedence rules.

    Precedence order (highest to lowest):
    1. Individual tool enable - if tool in enabled_tools, return True
    2. Individual tool disable - if tool in disabled_tools, return False
    3. Toolset enable - if tool's toolset in enabled_toolsets, return True
    4. Toolset disable - if tool's toolset in disabled_toolsets, return False
    5. Default - if any explicit enables are set, return False; otherwise return True

    Args:
        tool_name: The tool name as a ToolName enum
        enabled_tools: Set of explicitly enabled tools (highest precedence)
        disabled_tools: Set of explicitly disabled tools
        enabled_toolsets: Set of enabled toolsets
        disabled_toolsets: Set of disabled toolsets

    Returns:
        True if tool should be registered, False otherwise
    """
    # Precedence 1: Individual tool enable (highest)
    if tool_name in enabled_tools:
        return True

    # Precedence 2: Individual tool disable
    if tool_name in disabled_tools:
        return False

    # Get the tool's toolset
    tool_toolset = tool_to_toolset.get(tool_name)

    # Precedence 3 & 4: Toolset checks
    if tool_toolset:
        if tool_toolset in enabled_toolsets:
            return True
        if tool_toolset in disabled_toolsets:
            return False

    # Precedence 5: Fallback behavior (only when rules 1-4 don't apply)
    # This is the key mechanism that switches between denylist and allowlist modes:
    # - If NO enable configuration exists → return True (original default: enable all tools)
    # - If ANY enable configuration exists → return False (allowlist mode: disable unless explicitly allowed)
    # This preserves backward compatibility while enabling opt-in allowlist functionality
    return not bool(enabled_tools or enabled_toolsets)


def register_tools(
    dbt_mcp: FastMCP,
    tool_definitions: Sequence[GenericToolDefinition[ToolName]],
    *,
    disabled_tools: set[ToolName],
    enabled_tools: set[ToolName],
    enabled_toolsets: set[Toolset],
    disabled_toolsets: set[Toolset],
) -> None:
    return generic_register_tools(
        dbt_mcp,
        tool_definitions,
        disabled_tools=disabled_tools,
        enabled_tools=enabled_tools,
        enabled_toolsets=enabled_toolsets,
        disabled_toolsets=disabled_toolsets,
        tool_to_toolset=TOOL_TO_TOOLSET,
    )


def generic_register_tools[NameEnum: Enum](
    dbt_mcp: FastMCP,
    tool_definitions: Sequence[GenericToolDefinition[NameEnum]],
    *,
    disabled_tools: set[NameEnum],
    enabled_tools: set[NameEnum],
    enabled_toolsets: set[Toolset],
    disabled_toolsets: set[Toolset],
    tool_to_toolset: dict[NameEnum, Toolset],
) -> None:
    """Register tools with the MCP server using precedence-based enablement logic.

    Args:
        dbt_mcp: The FastMCP server instance
        tool_definitions: List of tool definitions to register
        exclude_tools: Tools to exclude (backward compatibility - converted to disabled_tools)
        enabled_tools: Set of explicitly enabled tools (precedence 1)
        enabled_toolsets: Set of enabled toolsets (precedence 3)
        disabled_toolsets: Set of disabled toolsets (precedence 4)
    """
    for tool_definition in tool_definitions:
        tool_name = tool_definition.get_name()
        if not should_register_tool(
            tool_name=tool_name,
            enabled_tools=enabled_tools,
            disabled_tools=disabled_tools,
            enabled_toolsets=enabled_toolsets,
            disabled_toolsets=disabled_toolsets,
            tool_to_toolset=tool_to_toolset,
        ):
            continue
        dbt_mcp.add_tool(
            fn=tool_definition.fn,
            name=tool_name.value,
            title=tool_definition.title,
            description=tool_definition.description,
            annotations=tool_definition.annotations,
            structured_output=tool_definition.structured_output,
        )



"""Toolset definitions and tool-to-toolset mappings.

This module defines the toolsets available in dbt-mcp and provides
a mapping from individual tools to their respective toolsets. This
enables toolset-level enablement/disablement configuration.
"""

from enum import Enum
from typing import Literal

from dbt_mcp.tools.tool_names import ToolName


class Toolset(Enum):
    SQL = "sql"
    SEMANTIC_LAYER = "semantic_layer"
    DISCOVERY = "discovery"
    DBT_CLI = "dbt_cli"
    ADMIN_API = "admin_api"
    DBT_CODEGEN = "dbt_codegen"
    DBT_LSP = "dbt_lsp"
    MCP_SERVER_METADATA = "mcp_server_metadata"


proxied_tools: set[
    Literal[
        ToolName.TEXT_TO_SQL,
        ToolName.EXECUTE_SQL,
        ToolName.GET_RELATED_MODELS,
        ToolName.SEARCH,
        ToolName.FUSION_COMPILE_SQL,
        ToolName.FUSION_GET_COLUMN_LINEAGE,
    ]
] = set(
    [
        ToolName.TEXT_TO_SQL,
        ToolName.EXECUTE_SQL,
        ToolName.GET_RELATED_MODELS,
        ToolName.SEARCH,
        ToolName.FUSION_COMPILE_SQL,
        ToolName.FUSION_GET_COLUMN_LINEAGE,
    ]
)

toolsets = {
    Toolset.SQL: {
        ToolName.TEXT_TO_SQL,
        ToolName.EXECUTE_SQL,
    },
    Toolset.SEMANTIC_LAYER: {
        ToolName.LIST_METRICS,
        ToolName.LIST_SAVED_QUERIES,
        ToolName.GET_DIMENSIONS,
        ToolName.GET_ENTITIES,
        ToolName.QUERY_METRICS,
        ToolName.GET_METRICS_COMPILED_SQL,
    },
    Toolset.DISCOVERY: {
        ToolName.GET_MART_MODELS,
        ToolName.GET_ALL_MODELS,
        ToolName.GET_MODEL_DETAILS,
        ToolName.GET_MODEL_PARENTS,
        ToolName.GET_MODEL_CHILDREN,
        ToolName.GET_MODEL_HEALTH,
        ToolName.GET_MODEL_PERFORMANCE,
        ToolName.GET_LINEAGE,
        ToolName.GET_ALL_SOURCES,
        ToolName.GET_SOURCE_DETAILS,
        ToolName.GET_EXPOSURES,
        ToolName.GET_EXPOSURE_DETAILS,
        ToolName.GET_RELATED_MODELS,
        ToolName.GET_MACRO_DETAILS,
        ToolName.GET_SEED_DETAILS,
        ToolName.GET_SEMANTIC_MODEL_DETAILS,
        ToolName.GET_SNAPSHOT_DETAILS,
        ToolName.GET_TEST_DETAILS,
        ToolName.SEARCH,
    },
    Toolset.DBT_CLI: {
        ToolName.BUILD,
        ToolName.COMPILE,
        ToolName.DOCS,
        ToolName.LIST,
        ToolName.PARSE,
        ToolName.RUN,
        ToolName.TEST,
        ToolName.SHOW,
        ToolName.GET_MODEL_LINEAGE_DEV,
        ToolName.GET_NODE_DETAILS_DEV,
    },
    Toolset.ADMIN_API: {
        ToolName.LIST_JOBS,
        ToolName.GET_JOB_DETAILS,
        ToolName.GET_PROJECT_DETAILS,
        ToolName.TRIGGER_JOB_RUN,
        ToolName.LIST_JOBS_RUNS,
        ToolName.GET_JOB_RUN_DETAILS,
        ToolName.CANCEL_JOB_RUN,
        ToolName.RETRY_JOB_RUN,
        ToolName.LIST_JOB_RUN_ARTIFACTS,
        ToolName.GET_JOB_RUN_ARTIFACT,
        ToolName.GET_JOB_RUN_ERROR,
    },
    Toolset.DBT_CODEGEN: {
        ToolName.GENERATE_SOURCE,
        ToolName.GENERATE_MODEL_YAML,
        ToolName.GENERATE_STAGING_MODEL,
    },
    Toolset.DBT_LSP: {
        ToolName.GET_COLUMN_LINEAGE,
        ToolName.FUSION_COMPILE_SQL,
        ToolName.FUSION_GET_COLUMN_LINEAGE,
    },
    Toolset.MCP_SERVER_METADATA: {
        ToolName.GET_MCP_SERVER_VERSION,
    },
}


# Mapping from individual tools to their toolsets (for precedence logic)
TOOL_TO_TOOLSET: dict[ToolName, Toolset] = {}
for toolset, tools in toolsets.items():
    for tool in tools:
        TOOL_TO_TOOLSET[tool] = toolset


def validate_tool_mapping() -> None:
    """Ensure all ToolName members are mapped to a toolset.

    Raises:
        ValueError: If any tools are not mapped to a toolset
    """
    unmapped = set(ToolName) - set(TOOL_TO_TOOLSET.keys())
    if unmapped:
        unmapped_names = [tool.value for tool in unmapped]
        raise ValueError(
            f"The following tools are not mapped to toolsets: {', '.join(unmapped_names)}"
        )


# Validate at import time to catch errors early
validate_tool_mapping()



from __future__ import annotations

from typing import Any

from pydantic import BaseModel

from dbt_mcp.oauth.token import (
    AccessTokenResponse,
    DecodedAccessToken,
    fetch_jwks_and_verify_token,
)


class DbtPlatformAccount(BaseModel):
    id: int
    name: str
    locked: bool
    state: int
    static_subdomain: str | None
    vanity_subdomain: str | None

    @property
    def host_prefix(self) -> str | None:
        if self.static_subdomain:
            return self.static_subdomain
        if self.vanity_subdomain:
            return self.vanity_subdomain
        return None


class DbtPlatformProject(BaseModel):
    id: int
    name: str
    account_id: int
    account_name: str


class DbtPlatformEnvironmentResponse(BaseModel):
    id: int
    name: str
    type: str | None = None
    deployment_type: str | None = None


class DbtPlatformEnvironment(BaseModel):
    id: int
    name: str
    deployment_type: str


class GetEnvironmentsRequest(BaseModel):
    account_id: int
    project_id: int


class SelectedProjectRequest(BaseModel):
    account_id: int
    project_id: int
    prod_environment_id: int | None = None


def dbt_platform_context_from_token_response(
    token_response: dict[str, Any], dbt_platform_url: str
) -> DbtPlatformContext:
    new_access_token_response = AccessTokenResponse(**token_response)
    decoded_claims = fetch_jwks_and_verify_token(
        new_access_token_response.access_token, dbt_platform_url
    )
    decoded_access_token = DecodedAccessToken(
        access_token_response=new_access_token_response,
        decoded_claims=decoded_claims,
    )
    return DbtPlatformContext(
        decoded_access_token=decoded_access_token,
    )


class DbtPlatformContext(BaseModel):
    decoded_access_token: DecodedAccessToken | None = None
    host_prefix: str | None = None
    dev_environment: DbtPlatformEnvironment | None = None
    prod_environment: DbtPlatformEnvironment | None = None
    account_id: int | None = None

    @property
    def user_id(self) -> int | None:
        return (
            int(self.decoded_access_token.decoded_claims["sub"])
            if self.decoded_access_token
            else None
        )

    def override(self, other: DbtPlatformContext) -> DbtPlatformContext:
        return DbtPlatformContext(
            dev_environment=other.dev_environment or self.dev_environment,
            prod_environment=other.prod_environment or self.prod_environment,
            decoded_access_token=other.decoded_access_token
            or self.decoded_access_token,
            host_prefix=other.host_prefix or self.host_prefix,
            account_id=other.account_id or self.account_id,
        )



import logging
from dataclasses import dataclass
from typing import Any

from mcp.server.fastmcp import FastMCP

from dbt_mcp.config.config_providers import (
    AdminApiConfig,
    ConfigProvider,
)
from dbt_mcp.dbt_admin.client import DbtAdminAPIClient
from dbt_mcp.dbt_admin.constants import STATUS_MAP, JobRunStatus
from dbt_mcp.dbt_admin.run_artifacts import ErrorFetcher, WarningFetcher
from dbt_mcp.prompts.prompts import get_prompt
from dbt_mcp.tools.definitions import dbt_mcp_tool
from dbt_mcp.tools.register import register_tools
from dbt_mcp.tools.tool_names import ToolName
from dbt_mcp.tools.toolsets import Toolset

logger = logging.getLogger(__name__)


@dataclass
class AdminToolContext:
    admin_client: DbtAdminAPIClient
    admin_api_config_provider: ConfigProvider[AdminApiConfig]

    def __init__(self, admin_api_config_provider: ConfigProvider[AdminApiConfig]):
        self.admin_api_config_provider = admin_api_config_provider
        self.admin_client = DbtAdminAPIClient(admin_api_config_provider)


@dbt_mcp_tool(
    description=get_prompt("admin_api/list_jobs"),
    title="List Jobs",
    read_only_hint=True,
    destructive_hint=False,
    idempotent_hint=True,
)
async def list_jobs(
    context: AdminToolContext,
    # TODO: add support for project_id in the future
    # project_id: Optional[int] = None,
    limit: int | None = None,
    offset: int | None = None,
) -> list[dict[str, Any]]:
    """List jobs in an account."""
    admin_api_config = await context.admin_api_config_provider.get_config()
    params = {}
    # if project_id:
    #     params["project_id"] = project_id
    if admin_api_config.prod_environment_id:
        params["environment_id"] = admin_api_config.prod_environment_id
    if limit:
        params["limit"] = limit
    if offset:
        params["offset"] = offset
    return await context.admin_client.list_jobs(admin_api_config.account_id, **params)


@dbt_mcp_tool(
    description=get_prompt("admin_api/get_job_details"),
    title="Get Job Details",
    read_only_hint=True,
    destructive_hint=False,
    idempotent_hint=True,
)
async def get_job_details(context: AdminToolContext, job_id: int) -> dict[str, Any]:
    """Get details for a specific job."""
    admin_api_config = await context.admin_api_config_provider.get_config()
    return await context.admin_client.get_job_details(
        admin_api_config.account_id, job_id
    )


@dbt_mcp_tool(
    description=get_prompt("admin_api/get_project_details"),
    title="Get Project Details",
    read_only_hint=True,
    destructive_hint=False,
    idempotent_hint=True,
)
async def get_project_details(
    context: AdminToolContext, project_id: int
) -> dict[str, Any]:
    """Get details for a specific project."""
    admin_api_config = await context.admin_api_config_provider.get_config()
    result = await context.admin_client.get_project_details(
        admin_api_config.account_id, project_id
    )
    for key in ("freshness_job", "docs_job", "group_permissions"):
        result.pop(key, None)
    return result


@dbt_mcp_tool(
    description=get_prompt("admin_api/trigger_job_run"),
    title="Trigger Job Run",
    read_only_hint=False,
    destructive_hint=False,
    idempotent_hint=False,
)
async def trigger_job_run(
    context: AdminToolContext,
    job_id: int,
    cause: str = "Triggered by dbt MCP",
    git_branch: str | None = None,
    git_sha: str | None = None,
    schema_override: str | None = None,
) -> dict[str, Any]:
    """Trigger a job run."""
    admin_api_config = await context.admin_api_config_provider.get_config()
    kwargs = {}
    if git_branch:
        kwargs["git_branch"] = git_branch
    if git_sha:
        kwargs["git_sha"] = git_sha
    if schema_override:
        kwargs["schema_override"] = schema_override
    return await context.admin_client.trigger_job_run(
        admin_api_config.account_id, job_id, cause, **kwargs
    )


@dbt_mcp_tool(
    description=get_prompt("admin_api/list_jobs_runs"),
    title="List Jobs Runs",
    read_only_hint=True,
    destructive_hint=False,
    idempotent_hint=True,
)
async def list_jobs_runs(
    context: AdminToolContext,
    job_id: int | None = None,
    status: JobRunStatus | None = None,
    limit: int | None = None,
    offset: int | None = None,
    order_by: str | None = None,
) -> list[dict[str, Any]]:
    """List runs in an account."""
    admin_api_config = await context.admin_api_config_provider.get_config()
    params: dict[str, Any] = {}
    if job_id:
        params["job_definition_id"] = job_id
    if status:
        status_id = STATUS_MAP[status]
        params["status"] = status_id
    if limit:
        params["limit"] = limit
    if offset:
        params["offset"] = offset
    if order_by:
        params["order_by"] = order_by
    return await context.admin_client.list_jobs_runs(
        admin_api_config.account_id, **params
    )


@dbt_mcp_tool(
    description=get_prompt("admin_api/get_job_run_details"),
    title="Get Job Run Details",
    read_only_hint=True,
    destructive_hint=False,
    idempotent_hint=True,
)
async def get_job_run_details(
    context: AdminToolContext,
    run_id: int,
) -> dict[str, Any]:
    """Get details for a specific job run."""
    admin_api_config = await context.admin_api_config_provider.get_config()
    return await context.admin_client.get_job_run_details(
        admin_api_config.account_id, run_id
    )


@dbt_mcp_tool(
    description=get_prompt("admin_api/cancel_job_run"),
    title="Cancel Job Run",
    read_only_hint=False,
    destructive_hint=False,
    idempotent_hint=False,
)
async def cancel_job_run(context: AdminToolContext, run_id: int) -> dict[str, Any]:
    """Cancel a job run."""
    admin_api_config = await context.admin_api_config_provider.get_config()
    return await context.admin_client.cancel_job_run(
        admin_api_config.account_id, run_id
    )


@dbt_mcp_tool(
    description=get_prompt("admin_api/retry_job_run"),
    title="Retry Job Run",
    read_only_hint=False,
    destructive_hint=False,
    idempotent_hint=False,
)
async def retry_job_run(context: AdminToolContext, run_id: int) -> dict[str, Any]:
    """Retry a failed job run."""
    admin_api_config = await context.admin_api_config_provider.get_config()
    return await context.admin_client.retry_job_run(admin_api_config.account_id, run_id)


@dbt_mcp_tool(
    description=get_prompt("admin_api/list_job_run_artifacts"),
    title="List Job Run Artifacts",
    read_only_hint=True,
    destructive_hint=False,
    idempotent_hint=True,
)
async def list_job_run_artifacts(context: AdminToolContext, run_id: int) -> list[str]:
    """List artifacts for a job run."""
    admin_api_config = await context.admin_api_config_provider.get_config()
    return await context.admin_client.list_job_run_artifacts(
        admin_api_config.account_id, run_id
    )


@dbt_mcp_tool(
    description=get_prompt("admin_api/get_job_run_artifact"),
    title="Get Job Run Artifact",
    read_only_hint=True,
    destructive_hint=False,
    idempotent_hint=True,
)
async def get_job_run_artifact(
    context: AdminToolContext, run_id: int, artifact_path: str, step: int | None = None
) -> Any:
    """Get a specific job run artifact."""
    admin_api_config = await context.admin_api_config_provider.get_config()
    return await context.admin_client.get_job_run_artifact(
        admin_api_config.account_id, run_id, artifact_path, step
    )


@dbt_mcp_tool(
    description=get_prompt("admin_api/get_job_run_error"),
    title="Get Job Run Error",
    read_only_hint=True,
    destructive_hint=False,
    idempotent_hint=True,
)
async def get_job_run_error(
    context: AdminToolContext,
    run_id: int,
    include_warnings: bool = False,
    warning_only: bool = False,
) -> dict[str, Any]:
    """Get focused error/warning information for a job run."""
    admin_api_config = await context.admin_api_config_provider.get_config()

    if warning_only:
        run_details = await context.admin_client.get_job_run_details(
            admin_api_config.account_id, run_id, include_logs=True
        )
        warning_fetcher = WarningFetcher(
            run_id, run_details, context.admin_client, admin_api_config
        )
        return await warning_fetcher.analyze_run_warnings()

    run_details = await context.admin_client.get_job_run_details(
        admin_api_config.account_id, run_id, include_logs=True
    )
    error_fetcher = ErrorFetcher(
        run_id, run_details, context.admin_client, admin_api_config
    )
    error_result = await error_fetcher.analyze_run_errors()

    if include_warnings:
        warning_fetcher = WarningFetcher(
            run_id, run_details, context.admin_client, admin_api_config
        )
        warning_result = await warning_fetcher.analyze_run_warnings()

        return {
            **error_result,
            "warnings": warning_result,
        }

    return error_result


ADMIN_TOOLS = [
    list_jobs,
    get_job_details,
    get_project_details,
    trigger_job_run,
    list_jobs_runs,
    get_job_run_details,
    cancel_job_run,
    retry_job_run,
    list_job_run_artifacts,
    get_job_run_artifact,
    get_job_run_error,
]


def register_admin_api_tools(
    dbt_mcp: FastMCP,
    admin_config_provider: ConfigProvider[AdminApiConfig],
    *,
    disabled_tools: set[ToolName],
    enabled_tools: set[ToolName],
    enabled_toolsets: set[Toolset],
    disabled_toolsets: set[Toolset],
) -> None:
    """Register dbt Admin API tools."""

    def bind_context() -> AdminToolContext:
        return AdminToolContext(admin_api_config_provider=admin_config_provider)

    register_tools(
        dbt_mcp,
        tool_definitions=[tool.adapt_context(bind_context) for tool in ADMIN_TOOLS],
        disabled_tools=disabled_tools,
        enabled_tools=enabled_tools,
        enabled_toolsets=enabled_toolsets,
        disabled_toolsets=disabled_toolsets,
    )



"""Common errors used across multiple tool types."""

from dbt_mcp.errors.base import ToolCallError


class InvalidParameterError(ToolCallError):
    """Exception raised when invalid or missing parameters are provided.

    This is a cross-cutting error used by multiple tool types.
    """

    pass



from typing import Any

import jwt
from jwt import PyJWKClient
from pydantic import BaseModel


class AccessTokenResponse(BaseModel):
    access_token: str
    refresh_token: str
    expires_in: int
    scope: str
    token_type: str
    expires_at: int


class DecodedAccessToken(BaseModel):
    access_token_response: AccessTokenResponse
    decoded_claims: dict[str, Any]


def fetch_jwks_and_verify_token(
    access_token: str, dbt_platform_url: str
) -> dict[str, Any]:
    jwks_url = f"{dbt_platform_url}/.well-known/jwks.json"
    jwks_client = PyJWKClient(jwks_url)
    signing_key = jwks_client.get_signing_key_from_jwt(access_token)
    claims = jwt.decode(
        access_token,
        signing_key.key,
        algorithms=["RS256"],
        options={"verify_aud": False},
    )
    return claims



"""dbt CLI tool errors."""

from dbt_mcp.errors.base import ToolCallError


class CLIToolCallError(ToolCallError):
    """Base exception for dbt CLI tool errors."""

    pass


class BinaryExecutionError(CLIToolCallError):
    """Exception raised when dbt binary execution fails."""

    pass



"""Binary detection and management for the dbt Language Server Protocol (LSP).

This module provides utilities to locate and validate the dbt LSP binary across
different operating systems and code editors (VS Code, Cursor, Windsurf). It handles
platform-specific paths and binary naming conventions.
"""

from enum import StrEnum
import os
from pathlib import Path
import platform
import subprocess
from dataclasses import dataclass

import logging

logger = logging.getLogger(__name__)

system = platform.system()
home = Path.home()


@dataclass
class LspBinaryInfo:
    """Information about a detected dbt LSP binary.

    Attributes:
        path: Full filesystem path to the LSP binary executable.
        version: Version string of the LSP binary.
    """

    path: str
    version: str


class CodeEditor(StrEnum):
    """Supported code editors that can install the dbt LSP.

    These editors use similar global storage patterns for VSCode extensions
    and can install the dbt Labs extension with the LSP binary.
    """

    CODE = "code"  # Visual Studio Code
    CURSOR = "cursor"  # Cursor editor
    WINDSURF = "windsurf"  # Windsurf editor


def get_storage_path(editor: CodeEditor) -> Path:
    """Get the storage path for dbt LSP binary based on editor and OS.

    Determines the platform-specific path where code editors store the dbt LSP
    binary. Follows standard conventions for each operating system and editor.

    Platform-specific paths:
        - Windows: %APPDATA%\\{editor}\\User\\globalStorage\\dbtlabsinc.dbt\\bin\\dbt-lsp
        - macOS: ~/Library/Application Support/{editor}/User/globalStorage/dbtlabsinc.dbt/bin/dbt-lsp
        - Linux: ~/.config/{editor}/User/globalStorage/dbtlabsinc.dbt/bin/dbt-lsp

    Args:
        editor: The code editor to get the storage path for.

    Returns:
        Path object pointing to the expected location of the dbt-lsp binary.

    Raises:
        ValueError: If the operating system is not supported (Windows, macOS, or Linux).

    Note:
        This function returns the expected path regardless of whether the binary
        actually exists at that location. Use Path.exists() to verify.
    """
    binary_name = "dbt-lsp"

    if system == "Windows":
        appdata = os.environ.get("APPDATA", home / "AppData" / "Roaming")
        base = Path(appdata) / editor.value
        binary_name = "dbt-lsp.exe"

    elif system == "Darwin":  # macOS
        base = home / "Library" / "Application Support" / editor.value

    elif system == "Linux":
        config_home = os.environ.get("XDG_CONFIG_HOME", home / ".config")
        base = Path(config_home) / editor.value

    else:
        raise ValueError(f"Unsupported OS: {system}")

    return Path(base, "User", "globalStorage", "dbtlabsinc.dbt", "bin", binary_name)


def dbt_lsp_binary_info(lsp_path: str | None = None) -> LspBinaryInfo | None:
    """Get dbt LSP binary information from a custom path or auto-detect it.

    Attempts to locate and validate the dbt LSP binary. If a custom path is provided,
    it will be validated first. If the custom path is invalid or not provided, the
    function will attempt to auto-detect the binary in standard editor locations.

    Args:
        lsp_path: Optional custom path to the dbt LSP binary. If provided, this
            path will be validated and used if it exists. If None or invalid,
            auto-detection will be attempted.

    Returns:
        LspBinaryInfo object containing the path and version of the found binary,
        or None if no valid binary could be found.

    Note:
        If a custom path is provided but invalid, a warning will be logged before
        falling back to auto-detection.
    """
    if lsp_path:
        logger.debug(f"Using custom LSP binary path: {lsp_path}")
        if Path(lsp_path).exists() and Path(lsp_path).is_file():
            version = get_lsp_binary_version(lsp_path)
            return LspBinaryInfo(path=lsp_path, version=version)
        logger.warning(
            f"Provided LSP binary path {lsp_path} does not exist or is not a file, falling back to detecting LSP binary"
        )
    return detect_lsp_binary()


def detect_lsp_binary() -> LspBinaryInfo | None:
    """Auto-detect dbt LSP binary in standard code editor locations.

    Searches through all supported code editors (VS Code, Cursor, Windsurf) to find
    an installed dbt LSP binary. Returns the first valid binary found.

    Returns:
        LspBinaryInfo object containing the path and version of the first found binary,
        or None if no binary is found in any of the standard locations.

    Note:
        The detection checks editors in the order defined by the CodeEditor enum.
        Debug logging is used to track the search process.
    """
    for editor in CodeEditor:
        path = get_storage_path(editor)
        logger.debug(f"Checking for LSP binary in {path}")
        if path.exists() and path.is_file():
            version = get_lsp_binary_version(path.as_posix())
            logger.debug(f"Found LSP binary in {path} with version {version}")
            return LspBinaryInfo(path=path.as_posix(), version=version)

    return None


def get_lsp_binary_version(path: str) -> str:
    """Extract the version string from a dbt LSP binary.

    Retrieves the version of the dbt LSP binary using one of two methods:
    1. If a .version file exists in the same directory as the binary, read from it
    2. Otherwise, execute the binary with --version flag

    Args:
        path: Full filesystem path to the dbt LSP binary.

    Returns:
        Version string of the binary (whitespace stripped).

    Raises:
        subprocess.SubprocessError: If the binary execution fails when .version file
            doesn't exist.

    Note:
        The .version file is expected to be in the same directory as the binary
        and should be named '.version'. This fallback behavior allows the function
        to work with both standard dbt-lsp installations and custom LSP binaries.
    """
    version_file = Path(path).parent / ".version"
    if version_file.exists():
        return version_file.read_text().strip()
    else:
        return subprocess.run(
            [path, "--version"], capture_output=True, text=True
        ).stdout.strip()



from pydantic import BaseModel, ConfigDict


class DbtProjectFlags(BaseModel):
    model_config = ConfigDict(extra="allow")
    send_anonymous_usage_stats: bool | None = None


class DbtProjectYaml(BaseModel):
    model_config = ConfigDict(extra="allow")
    flags: None | DbtProjectFlags = None



from collections.abc import Callable
from dataclasses import dataclass
from enum import Enum
from functools import partial
from typing import Any

from mcp.server.fastmcp.tools.base import Tool
from mcp.types import ToolAnnotations

from dbt_mcp.tools.injection import adapt_with_mapper
from dbt_mcp.tools.tool_names import ToolName


@dataclass
class GenericToolDefinition[NameEnum: Enum]:
    fn: Callable[..., Any]
    description: str
    name_enum: type[NameEnum]
    name: str | None = None
    title: str | None = None
    annotations: ToolAnnotations | None = None
    # We haven't strictly defined our tool contracts yet.
    # So we're setting this to False by default for now.
    structured_output: bool | None = False

    def get_name(self) -> NameEnum:
        return self.name_enum((self.name or self.fn.__name__).lower())

    def to_fastmcp_internal_tool(self) -> Tool:
        return Tool.from_function(
            fn=self.fn,
            name=self.name,
            title=self.title,
            description=self.description,
            annotations=self.annotations,
            structured_output=self.structured_output,
        )

    def adapt_context(
        self, context_mapper: Callable[..., Any]
    ) -> "GenericToolDefinition[NameEnum]":
        """
        Adapt the tool definition to accept a different context object.
        """
        return type(self)(
            fn=adapt_with_mapper(self.fn, context_mapper),
            description=self.description,
            name_enum=self.name_enum,
            name=self.name,
            title=self.title,
            annotations=self.annotations,
            structured_output=self.structured_output,
        )


@dataclass
class ToolDefinition(GenericToolDefinition[ToolName]):
    name_enum: type[ToolName] = ToolName


def generic_dbt_mcp_tool[NameEnum: Enum](
    description: str,
    name_enum: type[NameEnum],
    name: str | None = None,
    title: str | None = None,
    read_only_hint: bool = False,
    destructive_hint: bool = True,
    idempotent_hint: bool = False,
    open_world_hint: bool = True,
    structured_output: bool | None = False,
) -> Callable[[Callable], GenericToolDefinition[NameEnum]]:
    """Decorator to define a tool definition for dbt MCP"""

    def decorator(fn: Callable) -> GenericToolDefinition[NameEnum]:
        return GenericToolDefinition(
            fn=fn,
            description=description,
            name_enum=name_enum,
            name=name,
            title=title,
            annotations=ToolAnnotations(
                title=title,
                readOnlyHint=read_only_hint,
                destructiveHint=destructive_hint,
                idempotentHint=idempotent_hint,
                openWorldHint=open_world_hint,
            ),
            structured_output=structured_output,
        )

    return decorator


# Wrapper with ToolName pre-supplied for the common case
dbt_mcp_tool = partial(generic_dbt_mcp_tool, name_enum=ToolName)



from dataclasses import dataclass


@dataclass
class Misspelling:
    word: str
    similar_words: list[str]


def levenshtein(s1: str, s2: str) -> int:
    len_s1, len_s2 = len(s1), len(s2)
    dp = [[0] * (len_s2 + 1) for _ in range(len_s1 + 1)]

    for i in range(len_s1 + 1):
        dp[i][0] = i
    for j in range(len_s2 + 1):
        dp[0][j] = j

    for i in range(1, len_s1 + 1):
        for j in range(1, len_s2 + 1):
            cost = 0 if s1[i - 1] == s2[j - 1] else 1
            dp[i][j] = min(
                dp[i - 1][j] + 1,  # Deletion
                dp[i][j - 1] + 1,  # Insertion
                dp[i - 1][j - 1] + cost,  # Substitution
            )
    return dp[len_s1][len_s2]


def get_closest_words(
    target: str,
    words: list[str],
    top_k: int | None = None,
    threshold: int | None = None,
) -> list[str]:
    distances = [(word, levenshtein(target, word)) for word in words]

    # Filter by threshold if provided
    if threshold is not None:
        distances = [(word, dist) for word, dist in distances if dist <= threshold]

    # Sort by distance
    distances.sort(key=lambda x: x[1])

    # Limit by top_k if provided
    if top_k is not None:
        distances = distances[:top_k]

    return [word for word, _ in distances]


def get_misspellings(
    targets: list[str],
    words: list[str],
    top_k: int | None = None,
) -> list[Misspelling]:
    misspellings = []
    for target in targets:
        if target not in words:
            misspellings.append(
                Misspelling(
                    word=target,
                    similar_words=get_closest_words(
                        target=target,
                        words=words,
                        top_k=top_k,
                        threshold=max(1, len(target) // 2),
                    ),
                )
            )
    return misspellings



import logging
from dataclasses import dataclass

from mcp.server.fastmcp import FastMCP
from pydantic import Field

from dbt_mcp.config.config_providers import ConfigProvider, DiscoveryConfig
from dbt_mcp.discovery.client import (
    AppliedResourceType,
    ExposuresFetcher,
    LineageFetcher,
    LineageResourceType,
    MetadataAPIClient,
    ModelPerformanceFetcher,
    ModelsFetcher,
    PaginatedResourceFetcher,
    ResourceDetailsFetcher,
    SourcesFetcher,
)
from dbt_mcp.prompts.prompts import get_prompt
from dbt_mcp.tools.definitions import dbt_mcp_tool
from dbt_mcp.tools.register import register_tools
from dbt_mcp.tools.tool_names import ToolName
from dbt_mcp.tools.toolsets import Toolset

logger = logging.getLogger(__name__)

UNIQUE_ID_FIELD = Field(
    default=None,
    description="Fully-qualified unique ID of the resource. "
    "This will follow the format `<resource_type>.<package_name>.<resource_name>` "
    "(e.g. `model.analytics.stg_orders`). "
    "Strongly preferred over the `name` parameter for deterministic lookups.",
)
NAME_FIELD = Field(
    default=None,
    description="The name of the resource. "
    "This is not required if `unique_id` is provided. "
    "Only use name when `unique_id` is unknown.",
)
TYPES_FIELD = Field(
    default=None,
    description="List of resource types to include in lineage results. "
    "If not provided, includes all types. "
    "Valid types: Model, Source, Seed, Snapshot, Exposure, Metric, SemanticModel, SavedQuery, Test.",
)
DEPTH_FIELD = Field(default=5, description="The depth of the lineage graph to return.")


@dataclass
class DiscoveryToolContext:
    models_fetcher: ModelsFetcher
    exposures_fetcher: ExposuresFetcher
    sources_fetcher: SourcesFetcher
    resource_details_fetcher: ResourceDetailsFetcher
    lineage_fetcher: LineageFetcher
    model_performance_fetcher: ModelPerformanceFetcher

    def __init__(self, config_provider: ConfigProvider[DiscoveryConfig]):
        api_client = MetadataAPIClient(config_provider=config_provider)
        self.models_fetcher = ModelsFetcher(
            api_client=api_client,
            paginator=PaginatedResourceFetcher(
                api_client=api_client,
                edges_path=("data", "environment", "applied", "models", "edges"),
                page_info_path=("data", "environment", "applied", "models", "pageInfo"),
            ),
        )
        self.exposures_fetcher = ExposuresFetcher(
            api_client=api_client,
            paginator=PaginatedResourceFetcher(
                api_client=api_client,
                edges_path=("data", "environment", "definition", "exposures", "edges"),
                page_info_path=(
                    "data",
                    "environment",
                    "definition",
                    "exposures",
                    "pageInfo",
                ),
            ),
        )
        self.sources_fetcher = SourcesFetcher(
            api_client=api_client,
            paginator=PaginatedResourceFetcher(
                api_client,
                edges_path=("data", "environment", "applied", "sources", "edges"),
                page_info_path=(
                    "data",
                    "environment",
                    "applied",
                    "sources",
                    "pageInfo",
                ),
            ),
        )
        self.resource_details_fetcher = ResourceDetailsFetcher(api_client=api_client)
        self.lineage_fetcher = LineageFetcher(api_client=api_client)
        self.model_performance_fetcher = ModelPerformanceFetcher(
            api_client=api_client,
            resource_details_fetcher=self.resource_details_fetcher,
        )


@dbt_mcp_tool(
    description=get_prompt("discovery/get_mart_models"),
    title="Get Mart Models",
    read_only_hint=True,
    destructive_hint=False,
    idempotent_hint=True,
)
async def get_mart_models(context: DiscoveryToolContext) -> list[dict]:
    mart_models = await context.models_fetcher.fetch_models(
        model_filter={"modelingLayer": "marts"}
    )
    return [m for m in mart_models if m["name"] != "metricflow_time_spine"]


@dbt_mcp_tool(
    description=get_prompt("discovery/get_all_models"),
    title="Get All Models",
    read_only_hint=True,
    destructive_hint=False,
    idempotent_hint=True,
)
async def get_all_models(context: DiscoveryToolContext) -> list[dict]:
    return await context.models_fetcher.fetch_models()


@dbt_mcp_tool(
    description=get_prompt("discovery/get_model_details"),
    title="Get Model Details",
    read_only_hint=True,
    destructive_hint=False,
    idempotent_hint=True,
)
async def get_model_details(
    context: DiscoveryToolContext,
    name: str | None = NAME_FIELD,
    unique_id: str | None = UNIQUE_ID_FIELD,
) -> list[dict]:
    return await context.resource_details_fetcher.fetch_details(
        resource_type=AppliedResourceType.MODEL,
        unique_id=unique_id,
        name=name,
    )


@dbt_mcp_tool(
    description=get_prompt("discovery/get_model_parents"),
    title="Get Model Parents",
    read_only_hint=True,
    destructive_hint=False,
    idempotent_hint=True,
)
async def get_model_parents(
    context: DiscoveryToolContext,
    name: str | None = NAME_FIELD,
    unique_id: str | None = UNIQUE_ID_FIELD,
) -> list[dict]:
    return await context.models_fetcher.fetch_model_parents(name, unique_id)


@dbt_mcp_tool(
    description=get_prompt("discovery/get_model_children"),
    title="Get Model Children",
    read_only_hint=True,
    destructive_hint=False,
    idempotent_hint=True,
)
async def get_model_children(
    context: DiscoveryToolContext,
    name: str | None = NAME_FIELD,
    unique_id: str | None = UNIQUE_ID_FIELD,
) -> list[dict]:
    return await context.models_fetcher.fetch_model_children(name, unique_id)


@dbt_mcp_tool(
    description=get_prompt("discovery/get_model_health"),
    title="Get Model Health",
    read_only_hint=True,
    destructive_hint=False,
    idempotent_hint=True,
)
async def get_model_health(
    context: DiscoveryToolContext,
    name: str | None = NAME_FIELD,
    unique_id: str | None = UNIQUE_ID_FIELD,
) -> list[dict]:
    return await context.models_fetcher.fetch_model_health(name, unique_id)


@dbt_mcp_tool(
    description=get_prompt("discovery/get_model_performance"),
    title="Get Model Performance",
    read_only_hint=True,
    destructive_hint=False,
    idempotent_hint=True,
)
async def get_model_performance(
    context: DiscoveryToolContext,
    name: str | None = NAME_FIELD,
    unique_id: str | None = UNIQUE_ID_FIELD,
    num_runs: int = Field(
        default=1,
        description="Number of historical runs to return. Default is 1 (latest run only). "
        "Use values > 1 to analyze performance trends over time.",
        ge=1,
        le=100,
    ),
    include_tests: bool = Field(
        default=False,
        description="If True, include test execution history (name, status, executionTime) for each run. "
        "Useful for analyzing test performance alongside model execution. "
        "Default is False to reduce response size.",
    ),
) -> list[dict]:
    """Get model execution performance metrics from historical runs."""
    results = await context.model_performance_fetcher.fetch_performance(
        name=name,
        unique_id=unique_id,
        num_runs=num_runs,
        include_tests=include_tests,
    )

    return results


@dbt_mcp_tool(
    description=get_prompt("discovery/get_lineage"),
    title="Get Lineage",
    read_only_hint=True,
    destructive_hint=False,
    idempotent_hint=True,
)
async def get_lineage(
    context: DiscoveryToolContext,
    unique_id: str,
    types: list[LineageResourceType] | None = TYPES_FIELD,
    depth: int = DEPTH_FIELD,
) -> list[dict]:
    return await context.lineage_fetcher.fetch_lineage(
        unique_id=unique_id, types=types, depth=depth
    )


@dbt_mcp_tool(
    description=get_prompt("discovery/get_exposures"),
    title="Get Exposures",
    read_only_hint=True,
    destructive_hint=False,
    idempotent_hint=True,
)
async def get_exposures(context: DiscoveryToolContext) -> list[dict]:
    return await context.exposures_fetcher.fetch_exposures()


@dbt_mcp_tool(
    description=get_prompt("discovery/get_exposure_details"),
    title="Get Exposure Details",
    read_only_hint=True,
    destructive_hint=False,
    idempotent_hint=True,
)
async def get_exposure_details(
    context: DiscoveryToolContext,
    name: str | None = NAME_FIELD,
    unique_id: str | None = UNIQUE_ID_FIELD,
) -> list[dict]:
    return await context.resource_details_fetcher.fetch_details(
        resource_type=AppliedResourceType.EXPOSURE,
        unique_id=unique_id,
        name=name,
    )


@dbt_mcp_tool(
    description=get_prompt("discovery/get_all_sources"),
    title="Get All Sources",
    read_only_hint=True,
    destructive_hint=False,
    idempotent_hint=True,
)
async def get_all_sources(
    context: DiscoveryToolContext,
    source_names: list[str] | None = None,
    unique_ids: list[str] | None = None,
) -> list[dict]:
    return await context.sources_fetcher.fetch_sources(source_names, unique_ids)


@dbt_mcp_tool(
    description=get_prompt("discovery/get_source_details"),
    title="Get Source Details",
    read_only_hint=True,
    destructive_hint=False,
    idempotent_hint=True,
)
async def get_source_details(
    context: DiscoveryToolContext,
    name: str | None = NAME_FIELD,
    unique_id: str | None = UNIQUE_ID_FIELD,
) -> list[dict]:
    return await context.resource_details_fetcher.fetch_details(
        resource_type=AppliedResourceType.SOURCE,
        unique_id=unique_id,
        name=name,
    )


@dbt_mcp_tool(
    description=get_prompt("discovery/get_macro_details"),
    title="Get Macro Details",
    read_only_hint=True,
    destructive_hint=False,
    idempotent_hint=True,
)
async def get_macro_details(
    context: DiscoveryToolContext,
    name: str | None = NAME_FIELD,
    unique_id: str | None = UNIQUE_ID_FIELD,
) -> list[dict]:
    return await context.resource_details_fetcher.fetch_details(
        resource_type=AppliedResourceType.MACRO,
        unique_id=unique_id,
        name=name,
    )


@dbt_mcp_tool(
    description=get_prompt("discovery/get_seed_details"),
    title="Get Seed Details",
    read_only_hint=True,
    destructive_hint=False,
    idempotent_hint=True,
)
async def get_seed_details(
    context: DiscoveryToolContext,
    name: str | None = NAME_FIELD,
    unique_id: str | None = UNIQUE_ID_FIELD,
) -> list[dict]:
    return await context.resource_details_fetcher.fetch_details(
        resource_type=AppliedResourceType.SEED,
        unique_id=unique_id,
        name=name,
    )


@dbt_mcp_tool(
    description=get_prompt("discovery/get_semantic_model_details"),
    title="Get Semantic Model Details",
    read_only_hint=True,
    destructive_hint=False,
    idempotent_hint=True,
)
async def get_semantic_model_details(
    context: DiscoveryToolContext,
    name: str | None = NAME_FIELD,
    unique_id: str | None = UNIQUE_ID_FIELD,
) -> list[dict]:
    return await context.resource_details_fetcher.fetch_details(
        resource_type=AppliedResourceType.SEMANTIC_MODEL,
        unique_id=unique_id,
        name=name,
    )


@dbt_mcp_tool(
    description=get_prompt("discovery/get_snapshot_details"),
    title="Get Snapshot Details",
    read_only_hint=True,
    destructive_hint=False,
    idempotent_hint=True,
)
async def get_snapshot_details(
    context: DiscoveryToolContext,
    name: str | None = NAME_FIELD,
    unique_id: str | None = UNIQUE_ID_FIELD,
) -> list[dict]:
    return await context.resource_details_fetcher.fetch_details(
        resource_type=AppliedResourceType.SNAPSHOT,
        unique_id=unique_id,
        name=name,
    )


@dbt_mcp_tool(
    description=get_prompt("discovery/get_test_details"),
    title="Get Test Details",
    read_only_hint=True,
    destructive_hint=False,
    idempotent_hint=True,
)
async def get_test_details(
    context: DiscoveryToolContext,
    name: str | None = NAME_FIELD,
    unique_id: str | None = UNIQUE_ID_FIELD,
) -> list[dict]:
    return await context.resource_details_fetcher.fetch_details(
        resource_type=AppliedResourceType.TEST,
        unique_id=unique_id,
        name=name,
    )


DISCOVERY_TOOLS = [
    get_mart_models,
    get_all_models,
    get_model_details,
    get_model_parents,
    get_model_children,
    get_model_health,
    get_model_performance,
    get_lineage,
    get_exposures,
    get_exposure_details,
    get_all_sources,
    get_source_details,
    get_macro_details,
    get_seed_details,
    get_semantic_model_details,
    get_snapshot_details,
    get_test_details,
]


def register_discovery_tools(
    dbt_mcp: FastMCP,
    discovery_config_provider: ConfigProvider[DiscoveryConfig],
    *,
    disabled_tools: set[ToolName],
    enabled_tools: set[ToolName],
    enabled_toolsets: set[Toolset],
    disabled_toolsets: set[Toolset],
) -> None:
    def bind_context() -> DiscoveryToolContext:
        return DiscoveryToolContext(config_provider=discovery_config_provider)

    register_tools(
        dbt_mcp,
        tool_definitions=[tool.adapt_context(bind_context) for tool in DISCOVERY_TOOLS],
        disabled_tools=disabled_tools,
        enabled_tools=enabled_tools,
        enabled_toolsets=enabled_toolsets,
        disabled_toolsets=disabled_toolsets,
    )



GRAPHQL_QUERIES = {
    "metrics": """
query GetMetrics($environmentId: BigInt!, $search: String) {
  metricsPaginated(
    environmentId: $environmentId, search: $search
  ){
    items{
      name
      label
      description
      type
      config {
        meta
      }
    }
  }
}
    """,
    "dimensions": """
query GetDimensions($environmentId: BigInt!, $metrics: [MetricInput!]!, $search: String) {
  dimensionsPaginated(environmentId: $environmentId, metrics: $metrics, search: $search) {
    items {
      description
      name
      type
      queryableGranularities
      queryableTimeGranularities
      label
      config {
        meta
      }
    }
  }
}
    """,
    "entities": """
query GetEntities($environmentId: BigInt!, $metrics: [MetricInput!]!, $search: String) {
  entitiesPaginated(environmentId: $environmentId, metrics: $metrics, search: $search) {
    items {
      description
      name
      type
    }
  }
}
    """,
    "saved_queries": """
query GetSavedQueries($environmentId: BigInt!, $search: String) {
  savedQueriesPaginated(environmentId: $environmentId, search: $search) {
    items {
      name
      description
      label
      queryParams {
        metrics {
          name
        }
        groupBy {
          name
          grain
          datePart
        }
        where {
          whereSqlTemplate
        }
      }
    }
  }
}
    """,
}



#!/usr/bin/env python3
# This is a SUPER simplified version of the dbt manifest.json structure,
# only including the fields we need

from pydantic import BaseModel, Field


class Node(BaseModel):
    name: str


class Source(BaseModel):
    identifier: str


class Exposure(BaseModel):
    name: str


class Manifest(BaseModel):
    parent_map: dict[str, list[str]] = Field(default_factory=dict)
    child_map: dict[str, list[str]] = Field(default_factory=dict)
    nodes: dict[str, Node] = Field(default_factory=dict)
    sources: dict[str, Source] = Field(default_factory=dict)
    exposures: dict[str, Exposure] = Field(default_factory=dict)
